# FSGS - Common functionality for FS Game System.
# Copyright (C) 2013-2019  Frode Solheim <frode@solheim.dev>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
FSGS Game Driver for Commodore 64 (C64).

TODO:

* Handle V-SYNC
* Gamepad select button -> toggle status bar?
* Cleanup temp files
* Consider gamepad mappings for tape control
* Ability to select input devices, and control device type via database config
* Config key to set C64 model
* Gamepad button to open attack disk / tape menu?
* Handle save games properly (did C64 games save to tape? disk only?)
* Alt+S for screenshot and store screenshots to Screenshots dir
* Mouse support (need a use case / test game)

FIXME:

* FIXME: Positional keyboard support
  Keyboard: Error - Cannot load keymap `sdl_pos.vkm'.

"""

import json
import os
import shutil

from fsbc.system import macosx, windows
from fsgs.drivers.gamedriver import GameDriver
from fsgs.input.mapper import InputMapper
from fsgs.option import Option
from fsgs.platform import Platform
from fsgs.platforms.loader import SimpleLoader
from fsgs.util import sdl2

C64_MODEL_C64 = "c64"
C64_MODEL_C64C = "c64c"
C64_MODEL_C64C_1541_II = "c64c/1541-ii"
C64_JOYSTICK = {
    "type": "joystick",
    "description": "Joystick",
    "mapping_name": "c64",
}
C64_PORTS = [
    {
        "description": "Port 2",
        "types": [C64_JOYSTICK],
        "type_option": "c64_port_2_type",
        "device_option": "c64_port_2",
    },
    {
        "description": "Port 1",
        "types": [C64_JOYSTICK],
        "type_option": "c64_port_1_type",
        "device_option": "c64_port_1",
    },
]
C64_VIDEO_WIDTH = 384
C64_VIDEO_HEIGHT = 272

VICE_KEY_SET_A = 2
VICE_KEY_SET_B = 3
VICE_JOYSTICK = 4


class Commodore64Platform(Platform):
    PLATFORM_NAME = "Commodore 64"

    def __init__(self):
        super().__init__(Commodore64Loader, Commodore64ViceDriver)

    def driver(self, fsgc):
        return Commodore64ViceDriver(fsgc)

    def loader(self, fsgc):
        return Commodore64Loader(fsgc)


class Commodore64Loader(SimpleLoader):
    def load_files(self, values):
        file_list = json.loads(values["file_list"])
        # assert len(file_list) == 1
        for i, item in enumerate(file_list):
            _, ext = os.path.splitext(item["name"])
            ext = ext.upper()
            if ext in [".TAP", ".T64"]:
                if i == 0:
                    self.config["tape_drive_0"] = "sha1://{}/{}".format(
                        item["sha1"], item["name"]
                    )
                self.config["tape_image_{0}".format(i)] = (
                    "sha1://{}/{}".format(item["sha1"], item["name"])
                )
            elif ext in [".D64"]:
                if i == 0:
                    self.config["floppy_drive_0"] = "sha1://{}/{}".format(
                        item["sha1"], item["name"]
                    )
                self.config["floppy_image_{}".format(i)] = (
                    "sha1://{0}/{1}".format(item["sha1"], item["name"])
                )

    def load_extra(self, values):
        # FIXME: Replace with c64_model?
        self.config[Option.C64_MODEL] = values["model"]
        if not self.config[Option.C64_MODEL]:
            self.config[Option.C64_MODEL] = C64_MODEL_C64C
        self.config["c64_port_1_type"] = values["c64_port_1_type"]
        self.config["c64_port_2_type"] = values["c64_port_2_type"]
        # FIXME: Remove?
        self.config["model"] = ""


class Commodore64ViceDriver(GameDriver):
    PORTS = C64_PORTS

    def __init__(self, fsgs):
        super().__init__(fsgs)
        self.emulator.name = "x64sc-fs"
        self.helper = Commodore64Helper(self.options)

    def prepare(self):
        dot_vice_dir = os.path.join(self.home.path, ".vice")
        if os.path.exists(dot_vice_dir):
            shutil.rmtree(dot_vice_dir)
        os.makedirs(dot_vice_dir)
        dot_vice_dir = os.path.join(self.home.path, ".config", ".vice")
        if os.path.exists(dot_vice_dir):
            shutil.rmtree(dot_vice_dir)
        os.makedirs(dot_vice_dir)

        # noinspection SpellCheckingInspection
        joymap_file = self.temp_file("joymap.vjm").path
        with open(joymap_file, "w", encoding="UTF-8") as f:
            self.create_joymap_file(f)
        # noinspection SpellCheckingInspection
        if windows:
            # Not using normpath because os.sep can be "/" on MSYS2,
            # and Vice on Windows really requires backslashes here.
            joymap_file = joymap_file.replace("/", "\\")
        self.emulator.args.extend(["-joymap", joymap_file])

        hotkey_file = self.temp_file("hotkey.vkm").path
        with open(hotkey_file, "w", encoding="UTF-8") as f:
            self.create_hotkey_file(f)
        # noinspection SpellCheckingInspection
        if windows:
            hotkey_file = hotkey_file.replace("/", "\\")
        self.emulator.args.extend(["-hotkeyfile", hotkey_file])

        config_file = self.temp_file("vice.cfg").path
        with open(config_file, "w", encoding="UTF-8") as f:
            self.create_vice_cfg(f)
            self.configure_audio(f)
            self.configure_input(f)
            self.configure_video(f)
        self.emulator.args.extend(["-config", config_file])

        # self.emulator.args.extend(["-model", "c64"])
        if self.helper.model() == C64_MODEL_C64:
            self.set_model_name("Commodore C64")
            self.emulator.args.extend(["-model", "c64"])
        elif self.helper.model() == C64_MODEL_C64C_1541_II:
            self.set_model_name("Commodore C64C 1541-II")
            self.emulator.args.extend(["-model", "c64c"])
        else:
            self.set_model_name("Commodore C64C")
            self.emulator.args.extend(["-model", "c64c"])

        media_keys = ["floppy_drive_0", "tape_drive_0"]
        for i in range(20):
            media_keys.append("floppy_image_{0}".format(i))
            media_keys.append("tape_image_{0}".format(i))
        unique_uris = set()
        for key in media_keys:
            if self.options[key]:
                unique_uris.add(self.options[key])
        # media_dir = self.temp_dir("media")
        # VICE uses CWD as default directory for media files
        media_dir = self.cwd
        for file_uri in unique_uris:
            input_stream = self.fsgc.file.open(file_uri)
            game_file = os.path.join(media_dir.path, file_uri.split("/")[-1])
            with open(game_file, "wb") as f:
                f.write(input_stream.read())

        if self.options[Option.TAPE_DRIVE_0]:
            file_uri = self.options[Option.TAPE_DRIVE_0]
        else:
            file_uri = self.options[Option.FLOPPY_DRIVE_0]

        autostart_file = os.path.join(media_dir.path, file_uri.split("/")[-1])
        self.emulator.args.extend(["-autostart", autostart_file])

    def finish(self):
        pass

    def create_vice_cfg(self, f):
        f.write("[C64SC]\n")
        f.write("ConfirmOnExit=0\n")
        # f.write("KeepAspectRatio=1\n")
        f.write("VICIIDoubleScan=1\n")
        f.write("HwScalePossible=1\n")
        # 0 means SDL_FULLSCREEN_DESKTOP - 1 means SDL_FULLSCREEN
        f.write("VICIISDLFullscreenMode=0\n")
        f.write("AutostartWarp=0\n")
        f.write("\n")

        if self.helper.has_floppy_drive():
            f.write("Drive8Type=1541\n")
        else:
            # f.write("FileSystemDevice8=0\n")
            f.write("Drive8Type=None\n")
        f.write("\n")

        # Virtual device traps?
        # f.write("VirtualDevices=1\n")

        # print("set_media_options")
        # media_list = self.create_media_list()
        # print("sort media list")
        # # FIXME: SORT ON name (lowercase) ONLY, not whole path, because
        # # some files may have moved to another dir (temp dir)
        # media_list = sorted(media_list)
        # print(media_list)
        # if media_list[0].lower()[-4:] == ".crt":
        #     f.write("CartridgeFile=\"{path}\"\n".format(path=media_list[0]))
        #     f.write("CartridgeType={type}\n".format(type=0))
        #     f.write("CartridgeMode={mode}\n".format(mode=0))
        #     f.write("CartridgeReset={reset}\n".format(reset=1))
        # #    #self.args.extend(["-autostart", media_list[0]])
        # else:
        #     #f.write('AutostartPrgMode=2\n') # disk image
        #     #f.write('AutostartPrgDiskImage="{path}"\n'.format(
        #     #        path=media_list[0]))
        #     self.add_arg("-autostart", media_list[0])

        # # FIXME: Floppies?
        # print(media_list)
        # if media_list[0].lower()[-4:] == ".d64":
        #     f = self.context.temp.file('fliplist')
        #     f.write("# Vice fliplist file\n\n")
        #     f.write("UNIT 8\n")
        #     print("FLIP LIST:")
        #     # Make sure first disk is added to the end of the fliplist
        #     for floppy in (media_list[1:] + [media_list[:1]]):
        #         print("%s\n" % (floppy,))
        #         print
        #         f.write("%s\n" % (floppy,))
        #     f.close()
        #     self.add_arg("-flipname", self.context.temp.file("fliplist"))
        #
        # media_dir = os.path.dirname(media_list[0])
        # print(media_dir)
        # f.write("InitialDefaultDir=\"{dir}\"\n".format(dir=media_dir))
        # f.write("InitialTapeDir=\"{dir}\"\n".format(dir=media_dir))
        # f.write("InitialCartDir=\"{dir}\"\n".format(dir=media_dir))
        # f.write("InitialDiskDir=\"{dir}\"\n".format(dir=media_dir))
        # f.write("InitialAutostartDir=\"{dir}\"".format(dir=media_dir))

    # def vice_prepare_floppies(self):
    #     floppies = []
    #     #media_dir = os.path.dirname(self.context.game.file)
    #     #base_match = self.extract_match_name(os.path.basename(
    #     #        self.context.game.file))
    #     #for name in os.listdir(media_dir):
    #     #    dummy, ext = os.path.splitext(name)
    #     #    if ext.lower() not in ['.st']:
    #     #        continue
    #     #    match = self.extract_match_name(name)
    #     #    if base_match == match:
    #     #        floppies.append(os.path.join(media_dir, name))
    #     #        #floppies.append(name)
    #     if self.config["floppy_drive_0"]:
    #         floppies.append(self.config["floppy_drive_0"])
    #     return floppies

    def configure_audio(self, f):
        # Seems to be some issues with sound in VICE (buffers filling up,
        # slowing down the emulation?). Using defaults for now (uses big
        # buffers by default, only seems to delay the problem).
        # return
        audio_driver = self.options[Option.VICE_AUDIO_DRIVER]
        audio_driver = "sdl"
        f.write("SoundBufferSize={0}\n".format(50))
        if audio_driver:
            print("[VICE] Using audio driver", repr(audio_driver))
            f.write("SoundDeviceName={0}\n".format(audio_driver))
        # else:
        #     audio_driver = self.config.get("audio_driver", "")
        #     if audio_driver == "sdl":
        #         f.write("SoundDeviceName={0}\n".format(audio_driver))

        if self.use_audio_frequency():
            f.write("SoundSampleRate={0}\n".format(self.use_audio_frequency()))
        # default buffer size for vice is 100ms, that's far too much...
        # EDIT: Lowering the sound buffer size seems to cause FPS problems
        # Maybe due to buffer filling up and emu running slower (??)
        # f.write("SoundBufferSize={0}\n".format(50))
        if self.options[Option.FLOPPY_DRIVE_VOLUME] == 0:
            f.write("DriveSoundEmulation=0\n")
        else:
            # f.write("DriveSoundEmulationVolume=1200\n")
            f.write("DriveSoundEmulation=1\n")
        f.write("\n")

    def configure_input(self, f):
        # FIXME: Enable when ready
        # f.write("KeymapIndex=1\n")  # Use positional keys

        for i, port in enumerate(self.ports):
            vice_port = [2, 1, 3, 4][i]
            # vice_port = i + 1
            if port.device is None:
                vice_port_type = 0
            elif port.device.type == "joystick":
                vice_port_type = VICE_JOYSTICK
            elif port.device.type == "keyboard":
                vice_port_type = VICE_KEY_SET_A + i
                # assert False
            else:
                vice_port_type = 0
            f.write("JoyDevice{0}={1}\n".format(vice_port, vice_port_type))
            print(
                "[INPUT] Port",
                port.type_option,
                "VicePort",
                vice_port,
                port.type,
                port.device,
            )

            if port.device is None:
                continue
            if port.device.type != "keyboard":
                continue
            input_mapping = {
                "1": "KeySet{0}Fire".format(i + 1),
                "UP": "KeySet{0}North".format(i + 1),
                "DOWN": "KeySet{0}South".format(i + 1),
                "LEFT": "KeySet{0}West".format(i + 1),
                "RIGHT": "KeySet{0}East".format(i + 1),
            }
            mapper = ViceInputMapper(port, input_mapping)
            for key, value in mapper.items():
                f.write("{0}={1}\n".format(key, value))
        f.write("\n")

    def configure_video(self, f):
        # Enable the "Pepto" palette (http://www.pepto.de/projects/colorvic/)
        # f.write("VICIIPaletteFile=\"vice\"\n")
        palette_file = self.options[Option.C64_PALETTE]
        if not palette_file:
            # palette_file = "pepto-ntsc-sony"
            # palette_file = "vice"
            # palette_file = "community-colors"
            palette_file = "0"
        if palette_file != "0":
            f.write("VICIIExternalPalette=1\n")
            f.write('VICIIPaletteFile="{}"\n'.format(palette_file))

        f.write("VICIIAudioLeak=1\n")

        if self.effect() == self.CRT_EFFECT:
            f.write("VICIIFilter=1\n")
        elif self.effect() == self.DOUBLE_EFFECT:
            f.write("VICIIFilter=0\n")
        elif self.effect() == self.HQ2X_EFFECT:
            # HQ2X is not supported
            f.write("VICIIFilter=0\n")
        elif self.effect() == self.SCALE2X_EFFECT:
            f.write("VICIIFilter=2\n")
        else:
            f.write("VICIIFilter=0\n")
            f.write("VICIIDoubleSize=0\n")

        screen_w, screen_h = self.screen_size()

        if self.use_fullscreen():
            f.write("VICIIFullscreen=1\n")

        f.write("SDLWindowWidth={w}\n".format(w=960))
        f.write("SDLWindowHeight={h}\n".format(h=540))
        f.write("SDLCustomWidth={w}\n".format(w=screen_w))
        f.write("SDLCustomHeight={h}\n".format(h=screen_h))

        if self.scaling() == self.MAX_SCALING:
            f.write("VICIIHwScale=1\n")
        elif self.scaling() == self.INTEGER_SCALING:
            # FIXME: Does not support this yet, disabling scaling
            f.write("VICIIHwScale=0\n")
        else:
            f.write("VICIIHwScale=0\n")

        if self.stretching() == self.STRETCH_FILL_SCREEN:
            f.write("SDLGLAspectMode=0\n")
        elif self.stretching() == self.STRETCH_ASPECT:
            f.write("SDLGLAspectMode=2\n")
        else:
            f.write("SDLGLAspectMode=1\n")

        # if self.border() == self.LARGE_BORDER:
        #     f.write("VICIIBorderMode=1\n")
        # elif self.border() == self.SMALL_BORDER:
        #     # Value 4 is an FSGS extension in Vice-FS.
        #     f.write("VICIIBorderMode=4\n")
        # else:
        #     f.write("VICIIBorderMode=3\n")

        # Experimental "540 border mode"
        f.write("VICIIBorderMode=5\n")

        # # Disable scanlines in CRT mode
        # f.write("VICIIPALScanLineShade=1000\n")
        # Disable scanlines in CRT mode
        f.write("VICIIPALScanLineShade=850\n")

        # Can this be used to set fullscreen desktop, etc?
        # f.write("VICIISDLFullscreenMode = 1\n")

        f.write("\n")

    def create_joymap_file(self, f):
        # VICE joystick mapping file
        #
        # A joystick map is read in as patch to the current map.
        #
        # File format:
        # - comment lines start with '#'
        # - keyword lines start with '!keyword'
        # - normal line has 'joynum inputtype inputindex action'
        #
        # Keywords and their lines are:
        # '!CLEAR'    clear all mappings
        #
        # inputtype:
        # 0      axis
        # 1      button
        # 2      hat
        # 3      ball
        #
        # Note that each axis has 2 inputindex entries and each hat has 4.
        #
        # action [action_parameters]:
        # 0               none
        # 1 port pin      joystick (pin: 1/2/4/8/16 = u/d/l/r/fire)
        # 2 row col       keyboard
        # 3               map
        # 4               UI activate
        # 5 path&to&item  UI function
        #
        f.write("!CLEAR\n")
        for i, port in enumerate(self.ports):
            if port.device is None:
                continue
            if port.device.type != "joystick":
                continue
            if i == 0:
                input_mapping = {
                    "1": "1 1 16 ",
                    "UP": "1 1 1",
                    "DOWN": "1 1 2",
                    "LEFT": "1 1 4",
                    "RIGHT": "1 1 8",
                    "MENU": "4",
                }
            elif i == 1:
                input_mapping = {
                    "1": "1 0 16 ",
                    "UP": "1 0 1",
                    "DOWN": "1 0 2",
                    "LEFT": "1 0 4",
                    "RIGHT": "1 0 8",
                    "MENU": "4",
                }
            else:
                raise Exception("Invalid port")
            mapper = ViceInputMapper(port, input_mapping)
            for key, value in mapper.items():
                f.write("{0} {1}\n".format(value, key))

    def create_hotkey_file(self, f):
        # noinspection SpellCheckingInspection
        hotkeys = [
            (sdl2.SDLK_i, "Statusbar"),
            # (112, "Pause"),  # Mod+Pause
            # (113, "Quit emulator"),  # Mod+Q
            # ALT+S
            # (115, "Snapshot&Save snapshot image"),
            # (115, "Screenshot&Save PNG screenshot"),  # Mod+S
            (
                sdl2.SDLK_s,
                "Save media file&Create screenshot&Save PNG screenshot",
            ),
            (sdl2.SDLK_RETURN, "Video settings&Size settings&Fullscreen"),
            # (119, "Speed settings&Warp mode"),  # Mod+W
        ]
        if macosx:
            mod = 8  # Cmd key
        else:
            mod = 2  # Left alt key
        # f.write("!CLEAR\n")
        for key, action in hotkeys:
            f.write(
                "{0} {1}\n".format(sdl2.SDL_NUM_SCANCODES * mod + key, action)
            )


class ViceInputMapper(InputMapper):
    def axis(self, axis, positive):
        offset = 0 if positive else 1
        return "{0} 0 {1}".format(self.device.index, axis * 2 + offset)

    def hat(self, hat, direction):
        offset = {"left": 2, "right": 3, "up": 0, "down": 1}[direction]
        return "{0} 2 {1}".format(self.device.index, hat * 4 + offset)

    def button(self, button):
        return "{0} 1 {1}".format(self.device.index, button)

    def key(self, key):
        return "{0}".format(key.sdl_code)


class Commodore64Helper:
    def __init__(self, options):
        self.options = options

    def accuracy(self):
        try:
            accuracy = int(self.options.get(Option.ACCURACY, "1"))
        except ValueError:
            accuracy = 1
        return accuracy

    def has_floppy_drive(self):
        if self.model() == C64_MODEL_C64C_1541_II:
            return "1541-II"
        return None

    def model(self):
        if self.options[Option.C64_MODEL] == "c64":
            return C64_MODEL_C64
        elif self.options[Option.C64_MODEL] == "c64c":
            return C64_MODEL_C64C
        elif self.options[Option.C64_MODEL] == "c64c/1541-ii":
            return C64_MODEL_C64C_1541_II
        return C64_MODEL_C64C



from fsui.qt import QFileDialog
from fsui.qt.helpers import QParent

FILE = 0
FILES = 1
DIRECTORY = 2
NATIVE_DIALOGS = True


class FileDialog(QFileDialog):
    def __init__(
        self,
        parent=None,
        message="",
        directory="",
        file="",
        pattern="*.*",
        multiple=False,
        dir_mode=False,
    ):
        QFileDialog.__init__(self, QParent(parent), message)
        if directory:
            self.setDirectory(directory)
        if dir_mode:
            self.setFileMode(QFileDialog.FileMode.Directory)
        if multiple:
            self.setFileMode(QFileDialog.FileMode.ExistingFiles)
        # self.setWindowFlags()

    def __del__(self):
        print("FileDialog.__del__")

    def get_path(self):
        return self.get_paths()[0]

    def get_paths(self):
        return self.selectedFiles()

    def show_modal(self):
        result = self.exec()
        print("File dialog result is", result)
        return result

    # def show(self):
    #     return self.show_modal()


def pick_directory(parent=None, message="", directory=""):
    return pick_file(
        parent=parent, message=message, directory=directory, what=DIRECTORY
    )


def pick_file(parent=None, message="", directory="", what=FILE):
    if NATIVE_DIALOGS:
        if what == DIRECTORY:
            result = QFileDialog.getExistingDirectory(
                QParent(parent), message, directory
            )
        elif what == FILES:
            # return value is filenames, selected_filter
            result = QFileDialog.getOpenFileNames(
                QParent(parent), message, directory
            )
        else:
            # return value is filename, selected_filter
            result = QFileDialog.getOpenFileName(
                QParent(parent), message, directory
            )
        if isinstance(result, tuple):
            # PySide returns a tuple, path is first param
            result = result[0]
    else:
        dialog = QFileDialog(parent, message)
        if directory:
            dialog.setDirectory(directory)
        if what == DIRECTORY:
            dialog.setFileMode(QFileDialog.Directory)
        elif what == FILES:
            dialog.setFileMode(QFileDialog.ExistingFiles)
        modal_result = dialog.exec()
        if not modal_result:
            result = None
        elif what == FILES:
            result = dialog.selectedFiles()
        else:
            result = dialog.selectedFiles()[0]
        dialog.destroy()
    print("return file/dir dialog result:", result)
    return result


def pick_files(parent=None, message="", directory="", what=FILE):
    return pick_file(
        parent=parent, message=message, directory=directory, what=FILES
    )



from fsbc.desktop import open_url_in_browser

from .IconButton import IconButton


class HelpButton(IconButton):
    def __init__(self, parent, url):
        IconButton.__init__(self, parent, "16x16/information.png")
        self.url = url

    def on_activate(self):
        open_url_in_browser(self.url)



from __future__ import absolute_import
from __future__ import print_function

from .FSString import FSString

class FileName:
  root_path_aliases = (u'', u'/', u':')
  
  def __init__(self, name, is_intl=False, is_longname=False):
    # check that name is a FSString
    if not isinstance(name, FSString):
      raise ValueError("FileName's name must be a FSString")
    self.name = name
    self.is_intl = is_intl
    self.is_longname = is_longname
  
  def __str__(self):
    return self.name
  
  def __repr__(self):
    return self.name
  
  def is_root_path_alias(self):
    return self.name.get_unicode() in self.root_path_aliases
  
  def has_dir_prefix(self):
    return self.name.get_unicode().find("/") != -1

  def split_path(self):
    pc = self.name.get_unicode().split("/")
    p = []
    for path in pc:
      p.append(FileName(FSString(path), is_intl=self.is_intl, is_longname=self.is_longname))
    return p
  
  def get_dir_and_base_name(self):
    """Return portion after last slash '/' or the full name in unicode"""
    s = self.name.get_unicode()
    pos = s.rfind(u'/')
    if pos != -1:
      dir_name = s[:pos]
      file_name = s[pos+1:]
      if len(file_name) == 0:
        return FSString(dir_name), None
      else:
        return FSString(dir_name), FSString(file_name)
    else:
      return None, self.name
    
  def get_upper_ami_str(self):
    result = self.name.get_ami_str().upper()
    if self.is_intl:
      r = ""
      for i in xrange(len(result)):
        o = ord(result[i])
        if o >= 224 and o <= 254 and o != 247:
          r += chr(o - (ord('a')-ord('A')))
        else:
          r += chr(o)
      return r
    else:
      return result
  
  def is_valid(self):
    # check if path contains dir prefix components
    if self.has_dir_prefix():
        e = self.split_path()
        # empty path?
        if len(e) == 0:
            return False
        for p in e:
            if not p.is_valid():
                return False
        return True
    else:
        # single file name
        s = self.name.get_ami_str()
        # check for invalid chars
        for c in s:
          # o = ord(c)
          # if o == ':' or o == '/':
          # FIXME: FS
          if c == ':' or c == '/':
            return False
        # check max size
        if self.is_longname:
          if len(s) > 110:
              return False
        elif len(s) > 30:
            return False
    return True
  
  def hash(self, hash_size=72):
    up = self.get_upper_ami_str()
    h = len(up)
    for c in up:
      h = h * 13;
      h += ord(c)
      h &= 0x7ff
    h = h % hash_size
    return h

  def get_name(self):
    """Return file name string as a FSString."""
    return self.name

  def get_ami_str_name(self):
    return self.name.get_ami_str()
  
  def get_unicode_name(self):
    return self.name.get_unicode()



import os
import sqlite3
import threading

import fsbc.settings

global_database_lock = threading.Lock()


def log_query_plans():
    return fsbc.settings.get("log_query_plans") == "1"


def use_debug_cursor():
    return log_query_plans()


class ResetException(Exception):
    pass


class BaseDatabase(object):
    SENTINEL = "fae7671d-e232-4b71-b179-b3cd45995f92"

    thread_local = threading.local()

    @classmethod
    def get_path(cls):
        raise Exception("get_path not implemented")

    @classmethod
    def get_version(cls):
        raise Exception("get_version not implemented")

    @classmethod
    def get_reset_version(cls):
        return 0

    def __init__(self, sentinel):
        assert sentinel == self.SENTINEL
        self._internal_cursor = None
        self.connection = None

    def __del__(self):
        print("BaseDatabase.__del__")

    def __enter__(self):
        return self

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._internal_cursor = None
        if exc_type is None:
            self.commit()
        else:
            self.rollback()

    def init(self):
        if self.connection:
            return
        path = self.get_path()
        with global_database_lock:
            try:
                print("opening database", path)
                self.connection = sqlite3.connect(path)
                # noinspection PyPropertyAccess
                self.connection.row_factory = sqlite3.Row
                self.updated_database_if_needed()
            except ResetException:
                print("re-opening database", path)
                self.connection = sqlite3.connect(path)
                # noinspection PyPropertyAccess
                self.connection.row_factory = sqlite3.Row
                self.updated_database_if_needed()

    def internal_cursor(self):
        if self._internal_cursor is None:
            self._internal_cursor = self.cursor()
        return self._internal_cursor

    def cursor(self):
        if not self.connection:
            self.init()
        if use_debug_cursor():
            return DebuggingCursor(self.connection.cursor())
        return self.connection.cursor()

    def create_cursor(self):
        """ "
        :deprecated:
        """
        return self.cursor()

    def rollback(self):
        print("Database.rollback")
        if not self.connection:
            self.init()
        self.connection.rollback()

    def commit(self):
        print("BaseDatabase.commit")
        if not self.connection:
            self.init()
        self.connection.commit()

    def updated_database_if_needed(self):
        cursor = self.create_cursor()
        reset_version = self.get_reset_version()
        try:
            cursor.execute("SELECT version FROM metadata")
            version = cursor.fetchone()[0]
        except sqlite3.OperationalError:
            cursor.execute("CREATE TABLE metadata (version INTEGER)")
            cursor.execute("INSERT INTO metadata (version) VALUES (?)", (0,))
            self.connection.commit()
            version = 0

        if self.get_version() < version:
            raise Exception(
                "The database has been created by a newer version of this "
                "application. Have you started an older version?"
            )
        if 0 < version < self.get_reset_version():
            print(" -- resetting database -- ")
            self.connection.close()
            self.connection = None
            os.unlink(self.get_path())
            raise ResetException()

        if self.get_version() > version:
            self.update_database(
                max(version, reset_version - 1), self.get_version()
            )

    def update_database(self, old, new):
        for i in range(old + 1, new + 1):
            print("upgrading database to version", i)
            getattr(self, "update_database_to_version_{0}".format(i))()
            cursor = self.create_cursor()
            cursor.execute("UPDATE metadata SET version = ?", (i,))
            self.commit()


class DebuggingCursor(object):
    def __init__(self, cursor):
        self._cursor = cursor

    def __getattr__(self, item):
        return getattr(self._cursor, item)

    def __iter__(self):
        return self._cursor.__iter__()

    def execute(self, query, *args, **kwargs):
        print(query, args)
        if log_query_plans():
            self._cursor.execute(
                "EXPLAIN QUERY PLAN " + query, *args, **kwargs
            )
            for row in self._cursor:
                print(row[0], row[1], row[2], row[3])
        return self._cursor.execute(query, *args, **kwargs)



import fsui
from launcher.i18n import gettext
from launcher.settings.settings_page import SettingsPage


class WHDLoadSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("settings", "pkg:workspace")
        # gettext("WHDLoad Settings")
        title = gettext("WHDLoad")
        subtitle = gettext("Options for WHDLoad support in FS-UAE Launcher")
        self.add_header(icon, title, subtitle)

        label = fsui.MultiLineLabel(
            self,
            gettext(
                "The following options only apply when you use the automatic "
                "WHDLoad support in FS-UAE Launcher, for example in relation with "
                "the online game database."
            ),
            640,
        )
        self.layout.add(label, fill=True, margin_top=0)

        self.add_option("whdload_preload")
        self.add_option("whdload_splash_delay")
        self.add_option("whdload_quit_key")

        # label = fsui.Label(
        #     self, gettext("Directory for WHDLoad.key file (if you have it):"))
        # self.layout.add(label, margin_top=10)
        # hor_layout = fsui.HorizontalLayout()
        # self.layout.add(hor_layout, margin_top=4, fill=True)
        # label = fsui.Label(
        #     self, FSGSDirectories.get_base_dir())
        # hor_layout.add_spacer(0, expand=True)
        # hor_layout.add(label)



import fsui
from fsgs.platform import Platform
from launcher.option import Option
from launcher.settings.option_ui import OptionUI


class PlatformSettingsDialog(fsui.Window):
    @classmethod
    def open(cls, parent, platform):
        # return fsui.open_window_instance(cls, parent)
        dialog = cls(parent, platform)
        dialog.show()
        return dialog

    def __init__(self, parent, platform):
        title = "Platform Settings: " + platform.upper()
        super().__init__(parent, title)
        # self.layout = fsui.VerticalLayout()
        buttons, layout = fsui.DialogButtons.create_with_layout(self)
        # if self.window.theme.has_close_buttons:
        # buttons.create_close_button()

        # self.layout.padding = 10

        for option in self.option_list_for_platform(platform):
            self.add_option(layout, option)

        self.set_size((600, 400))

    def add_option(self, layout, option, platforms=None, text=""):
        panel = fsui.Panel(self)
        panel.layout = fsui.VerticalLayout()
        panel.layout.add(
            OptionUI.create_group(
                panel, option, text, thin=False, help_button=True
            ),
            fill=True,
        )
        layout.add(panel, fill=True, margin_bottom=10)

    @staticmethod
    def option_list_for_platform(platform):
        options = []
        if platform == Platform.ARCADE:
            # options.append(Option.ARCADE_DATABASE)
            options.append(Option.MAME_ARTWORK)
            options.extend(mame_options)
        elif platform == Platform.C64:
            # options.append(Option.C64_DATABASE)
            options.append(Option.C64_PALETTE)
            options.append(Option.VICE_AUDIO_DRIVER)
        elif platform == Platform.GBA:
            # options.append(Option.NES_DATABASE)
            # options.append(Option.GBA_EMULATOR)
            options.append(Option.GBA_GAMMA)
            options.extend(mednafen_options)
        elif platform == Platform.NEOGEO:
            # options.append(Option.NEOGEO_DATABASE)
            options.extend(mame_options)
        elif platform == Platform.NES:
            # options.append(Option.NES_DATABASE)
            options.append(Option.NES_EMULATOR)
            options.extend(mednafen_options)
            options.extend(retroarch_options)
        elif platform == Platform.PSX:
            # options.append(Option.PSX_DATABASE)
            options.append(Option.PSX_PRELOAD)
            options.extend(mednafen_options)
            options.append(Option.MEDNAFEN_DEINTERLACER)
            options.append(Option.MEDNAFEN_TEMPORAL_BLUR)
        elif platform == Platform.SMD:
            # options.append(Option.SMD_DATABASE)
            options.append(Option.SMD_EMULATOR)
            options.extend(mednafen_options)
            options.extend(retroarch_options)
        elif platform == Platform.SNES:
            # options.append(Option.SNES_DATABASE)
            options.extend(mednafen_options)
        return options


mame_options = []

mednafen_options = [Option.MEDNAFEN_AUDIO_DRIVER, Option.MEDNAFEN_AUDIO_BUFFER]

retroarch_options = [Option.RETROARCH_AUDIO_BUFFER]



import sys


class FSString:
  """Simple string class that allows to manage strings encoded in Latin-1 used for the Amiga FS.
     It stores the string internally as a python UTF-8 string but allows to convert to Amiga format.
  """
  
  def __init__(self, txt, encoding="Latin-1"):
    """Init the string. Either with a unicode string or with a 8-Bit string.
       If the latter is given then the "encoding" flag determines the encoding.
    """
    if sys.version_info[0] == 3 and type(txt) == str:
      self.txt = txt
    elif type(txt) == unicode:
      self.txt = txt
    elif type(txt) == str:
      self.txt = txt.decode(encoding)
    else:
      raise ValueError("FSString must be str or unicode!")
      
  def __repr__(self):
    return self.__str__()
  
  def __str__(self):
    return self.__unicode__().encode("UTF-8")
    
  def __unicode__(self):
    return self.txt
  
  def get_unicode(self):
    return self.txt
  
  def get_ami_str(self):
    return self.txt.encode("Latin-1")
  


from __future__ import absolute_import
from __future__ import print_function

import os.path
import sys

from .Block import Block
import amitools.fs.DosType as DosType

if sys.version_info[0] == 3:
  xrange = range


class BootBlock(Block):  
  def __init__(self, blkdev, blk_num=0):
    Block.__init__(self, blkdev, blk_num)
    self.dos_type = None
    self.got_root_blk = None
    self.got_chksum = 0
    self.calc_chksum = 0
    self.boot_code = None
    self.num_extra = self.blkdev.bootblocks - 1
    self.max_boot_code = self.blkdev.bootblocks * self.blkdev.block_bytes - 12
    self.extra_blks = []
  
  def create(self, dos_type=DosType.DOS0, root_blk=None, boot_code=None):
    Block.create(self)
    self._create_data()
    self.dos_type = dos_type    
    self.valid_dos_type = True
    # root blk
    self.calc_root_blk = int(self.blkdev.num_blocks / 2)
    if root_blk != None:
      self.got_root_blk = root_blk
    else:
      self.got_root_blk = self.calc_root_blk      
    # create extra blks
    self.extra_blks = []
    for i in xrange(self.num_extra):
      b = Block(self.blkdev, self.blk_num + 1 + i)
      b._create_data()
      self.extra_blks.append(b)
    # setup boot code
    return self.set_boot_code(boot_code)
  
  def set_boot_code(self, boot_code):
    if boot_code != None:
      if len(boot_code) <= self.max_boot_code:
        self.boot_code = boot_code
        self.valid = True
      else:
        self.valid = False
    else:
      self.boot_code = None
      self.valid = True   
    return self.valid
  
  def _calc_chksum(self):
    all_blks = [self] + self.extra_blks
    n = self.blkdev.block_longs
    chksum = 0
    for blk in all_blks:
      for i in xrange(n):
        if i != 1: # skip chksum
          chksum += blk._get_long(i)
          if chksum > 0xffffffff:
            chksum += 1
            chksum &= 0xffffffff
    return (~chksum) & 0xffffffff
  
  def read(self):
    self._read_data()
    # read extra boot blocks
    self.extra_blks = []
    for i in xrange(self.num_extra):
      b = Block(self.blkdev, self.blk_num + 1 + i)
      b._read_data()
      self.extra_blks.append(b)
      
    self.dos_type = self._get_long(0)
    self.got_chksum = self._get_long(1)
    self.got_root_blk = self._get_long(2)
    self.calc_chksum = self._calc_chksum()
    # calc position of root block
    self.calc_root_blk = int(self.blkdev.num_blocks / 2)
    # check validity
    self.valid_chksum = self.got_chksum == self.calc_chksum
    self.valid_dos_type = DosType.is_valid(self.dos_type)
    self.valid = self.valid_dos_type

    # look for boot_code
    if self.valid:
      self.read_boot_code()
    return self.valid
  
  def read_boot_code(self):
    boot_code = self.data[12:]
    for blk in self.extra_blks:
      boot_code += blk.data.raw
    # remove nulls at end
    pos = len(boot_code) - 4
    while pos > 0:
      tag = boot_code[pos:pos+3]
      if tag != 'DOS' and ord(boot_code[pos])!=0:
        pos += 4
        break
      pos -= 4
    # something left
    if pos > 0:
      boot_code = boot_code[:pos]
      self.boot_code = boot_code
  
  def write(self):
    self._create_data()
    self._put_long(0, self.dos_type)
    self._put_long(2, self.got_root_blk)

    if self.boot_code != None:
      self.write_boot_code()
      self.calc_chksum = self._calc_chksum()
      self._put_long(1, self.calc_chksum)
      self.valid_chksum = True
    else:
      self.calc_chksum = 0
      self.valid_chksum = False

    self._write_data()
  
  def write_boot_code(self):
    n = len(self.boot_code)
    bb = self.blkdev.block_bytes
    first_size = bb - 12
    boot_code = self.boot_code
    # spans more blocks
    if n > first_size:
      extra = boot_code[first_size:]
      boot_code = boot_code[:first_size]
      # write extra blocks
      pos = 0
      off = 0
      n -= first_size
      while n > 0:
        num = n
        if num > bb:
          num = bb
        self.extra_blks[pos].data[:num] = extra[off:off+num]
        self.extra_blks[pos]._write_data()
        off += num
        pos += 1
        n -= num
      # use this for first block
      n = first_size
        
    # embed boot code in boot block
    self.data[12:12+n] = boot_code    
  
  def dump(self):
    print("BootBlock(%d):" % self.blk_num)
    print(" dos_type:  0x%08x %s (valid: %s) is_ffs=%s is_intl=%s is_dircache=%s" \
      % (self.dos_type, DosType.num_to_tag_str(self.dos_type), 
         self.valid_dos_type, 
         DosType.is_ffs(self.dos_type), 
         DosType.is_intl(self.dos_type),
         DosType.is_dircache(self.dos_type)))
    print(" root_blk:  %d (got %d)" % (self.calc_root_blk, self.got_root_blk))
    print(" chksum:    0x%08x (got) 0x%08x (calc) -> bootable: %s" % (self.got_chksum, self.calc_chksum, self.valid_chksum))
    print(" valid:     %s" % self.valid)
    if self.boot_code != None:
        print(" boot_code: %d bytes" % len(self.boot_code))

  def get_boot_code_dir(self):
    my_dir = os.path.dirname(__file__)
    bc_dir = os.path.join(my_dir, "bootcode")
    if os.path.exists(bc_dir):
      return bc_dir
    else:
      return None

    


import traceback
import queue

from fsui.res import gettext
from fsui.qt.qt import (
    init_qt,
    QGuiApplication,
    QEvent,
    QObject,
    QCoreApplication,
    QTimer,
    QMessageBox,
    QCursor,
)
from fsui.qt.qt import *
from fsui.qt.helpers import QParent


def get_screen_size():
    init_qt()
    screen = QGuiApplication.primaryScreen()
    geometry = screen.geometry()
    size = geometry.width(), geometry.height()
    print("using screen size", size)
    return size


def get_mouse_position():
    # noinspection PyArgumentList
    pos = QCursor.pos()
    return pos.x(), pos.y()


class CustomEvent(QEvent):
    def __init__(self):
        QEvent.__init__(self, QEvent.Type.User)


class EventHandler(QObject):
    def __init__(self):
        QObject.__init__(self)
        self.queue = queue.Queue()

    def customEvent(self, a0):
        while True:
            try:
                function, args, kwargs = self.queue.get_nowait()
            except queue.Empty:
                break
            try:
                function(*args, **kwargs)
            except Exception:
                # log.warn("callback event failed: %r %r",
                #         self.callback, self.args, exc_info=True)
                print("-- callback exception --")
                traceback.print_exc()

    def post_callback(self, function, *args, **kwargs):
        self.queue.put((function, args, kwargs))
        QCoreApplication.instance().postEvent(self, CustomEvent())


event_handler = EventHandler()


def call_after(function, *args, **kwargs):
    event_handler.post_callback(function, *args, **kwargs)


def call_later(duration, function, *args, **kwargs):
    # print("FIXME: call_later", duration, function)
    # raise NotImplementedError()
    # QApplication.instance().
    def timer_callback():
        function(*args, **kwargs)

    QTimer.singleShot(duration, timer_callback)


def show_error(message, title=None, parent=None):
    if not title:
        title = gettext("An Error Occurred")
    # QErrorMessage().showMessage(message)
    # message_box = QMessageBox()
    # message_box.setIcon(QMessageBox.Critical)
    # message_box.setText(message)
    # message_box.exec()
    QMessageBox.critical(QParent(parent), title, message)


def show_warning(message, title=None, parent=None):
    if not title:
        title = gettext("Warning")
    QMessageBox.warning(QParent(parent), title, message)


def error_function(title):
    def error_function_2(message):
        show_error(message, title)

    return error_function_2



from fsgs.drivers.messdriver import MessDriver


class MessCpcDriver(MessDriver):
    CONTROLLER = {
        "type": "controller",
        "description": "Controller",
        "mapping_name": "cpc",
    }

    PORTS = [
        {"description": "1st Controller", "types": [CONTROLLER]},
        {"description": "2nd Controller", "types": [CONTROLLER]},
    ]

    def mess_configure(self):
        self.mess_configure_floppies(["flop1", "flop2"])
        if self.config["command"]:
            # pass
            # self.inject_fake_input_string(60, "{0}\n".format(
            #     self.config["command"]))
            # self.args.extend(["-autoboot_delay", ""])
            self.args.extend(
                ["-autoboot_command", self.config["command"] + r"\n"]
            )

    def mess_full_keyboard(self):
        return True

    def mess_input_mapping(self, port):
        return {
            "UP": "P#_JOYSTICK_UP",
            "DOWN": "P#_JOYSTICK_DOWN",
            "LEFT": "P#_JOYSTICK_LEFT",
            "RIGHT": "P#_JOYSTICK_RIGHT",
            "1": "P#_BUTTON1",
            "2": "P#_BUTTON2",
        }

    def get_game_refresh_rate(self):
        return 50.080128

    def mess_romset(self):
        model = self.config["model"].lower()
        if model in ["6128", "cpc6128"]:
            return "cpc6128", CPC6128_ROMS
        elif model in ["664", "cpc664"]:
            return "cpc664", CPC664_ROMS
        elif model in ["464", "cpc464", ""]:
            return "cpc464", CPC464_ROMS
        raise Exception("Unexpected CPC model " + repr(model))


# noinspection SpellCheckingInspection
CPC464_ROMS = {
    "56d39c463da60968d93e58b4ba0e675829412a20": "cpc464.rom",
    "39102c8e9cb55fcc0b9b62098780ed4a3cb6a4bb": "cpcados.rom",
}

# noinspection SpellCheckingInspection
CPC664_ROMS = {
    "073a7665527b5bd8a148747a3947dbd3328682c8": "cpc664.rom",
    "39102c8e9cb55fcc0b9b62098780ed4a3cb6a4bb": "cpcados.rom",
}

# noinspection SpellCheckingInspection
CPC6128_ROMS = {
    "5977adbad3f7c1e0e082cd02fe76a700d9860c30": "cpc6128.rom",
    "39102c8e9cb55fcc0b9b62098780ed4a3cb6a4bb": "cpcados.rom",
}



from fsgs.knownfiles import KnownFile
from fsgs.platform import Platform
from fsgs.platforms.loader import CDPlatformLoader
from fsgs.platforms.turbografx16 import TurboGrafx16MednafenDriver

TGCD_PLATFORM_ID = "tgcd"
TGCD_PLATFORM_NAME = "TurboGrafx-CD"
TGCD_CONTROLLER = {
    "type": "gamepad",
    "description": "Gamepad",
    "mapping_name": "turbografx16",
}
SYSCARD3_PCE = KnownFile(
    "1b4c260326d905bc718812dad0f68089977f427b",
    TGCD_PLATFORM_ID,
    "syscard3.pce",
)


class TurboGrafxCDPlatform(Platform):
    PLATFORM_ID = TGCD_PLATFORM_ID
    PLATFORM_NAME = TGCD_PLATFORM_NAME

    def __init__(self):
        super().__init__()

    def driver(self, fsgc):
        return TurboGrafxCDMednafenDriver(fsgc)

    def loader(self, fsgc):
        return TurboGrafxCDLoader(fsgc)


class TurboGrafxCDLoader(CDPlatformLoader):
    pass


class TurboGrafxCDMednafenDriver(TurboGrafx16MednafenDriver):
    PORTS = [
        {"description": "Input Port 1", "types": [TGCD_CONTROLLER]},
        {"description": "Input Port 2", "types": [TGCD_CONTROLLER]},
    ]

    def __init__(self, fsgc):
        super().__init__(fsgc)

    # def game_video_par(self):
    #     return (4 / 3) / (288 / 232)
    #
    # def game_video_size(self):
    #     return 288 * 2, 232 * 2

    def mednafen_system_prefix(self):
        return "pce"

    def prepare(self):
        # self.emulator.args.extend(
        #     ["-pce.enable", "0", "-pce_fast.enable", "0"])
        # self.emulator.args.extend(["-force_module", "pce_fast"])
        # self.emulator.args.extend(["-force_module", "pce"])
        super().prepare()
        self.prepare_mednafen_bios(SYSCARD3_PCE, "syscard3.pce")
        self.prepare_mednafen_cd_images()

    def get_game_file(self, config_key="cartridge_slot"):
        return None



import os
import sys
import time
import weakref

import fsgs as fsgs_module
import fstd.desktop
import fsui
import launcher.ui
from fsbc.application import Application
from fsbc.util import unused
from fsgs.context import default_context
from fsgs.ogd.locker import is_locker_enabled
from launcher.i18n import gettext
from launcher.implicit_handler import ImplicitConfigHandler
from launcher.launcher_config import LauncherConfig
from launcher.launcher_settings import LauncherSettings
from launcher.launcher_signal import LauncherSignal
from launcher.netplay.netplay_dialog import NetplayDialog
from launcher.option import Option
from launcher.panels.additionalconfigpanel import AdditionalConfigPanel
from launcher.panels.cdpanel import CDPanel
from launcher.panels.expansionspanel import ExpansionsPanel
from launcher.panels.floppiespanel import FloppiesPanel
from launcher.panels.harddrivespanel import HardDrivesPanel
from launcher.panels.inputpanel import InputPanel
from launcher.panels.mainpanel import MainPanel
from launcher.panels.quicksettingspanel import QuickSettingsPanel
from launcher.panels.romrampanel import RomRamPanel
from launcher.ui.aboutdialog import AboutDialog
from launcher.ui.book import Book
from launcher.ui.bottombar.BottomPanel import BottomPanel
from launcher.ui.bottombar.GameInfoPanel import GameInfoPanel
from launcher.ui.bottombar.ScreenshotsPanel import ScreenshotsPanel
from launcher.ui.config.configscrollarea import ConfigScrollArea
from launcher.ui.ConfigurationsPanel import ConfigurationsPanel
from launcher.ui.Constants import Constants
from launcher.ui.kickstartimportdialog import KickstartImportDialog
from launcher.ui.launch import LaunchGroup
from launcher.ui.skin import Skin
from launcher.ui.statusbar.StatusBar import StatusBar
from launcher.ui.WindowWithTabs import WindowWithTabs
from launcher.update_manager import UpdateManager
from workspace.apps.adf_creator_app import ADFCreatorWindow
from workspace.apps.hdf_creator_app import HDFCreatorWindow
from workspace.apps.locker_uploader import LockerUploaderWindow
from workspace.apps.login import LoginWindow
from workspace.apps.logout import LogoutWindow
from workspace.apps.refresh import RefreshWindow

USE_MAIN_MENU = 1
QUICK_SETTINGS_WIDTH = 200
CONFIG_BROWSER_WIDTH = 200
GAME_INFO_PANEL_MIN_WIDTH = 500

SCREENSHOT_WIDTH = 212
SCREENSHOT_SPACING = 20
MARGIN = 20
CONFIGURATIONS_PANEL_WIDTH_1 = (
    SCREENSHOT_WIDTH * 2 + SCREENSHOT_SPACING * 1 + MARGIN
)
CONFIGURATIONS_PANEL_WIDTH_2 = (
    SCREENSHOT_WIDTH * 3 + SCREENSHOT_SPACING * 2 + MARGIN
)
CONFIGURATIONS_PANEL_WIDTH_THRESHOLD_2 = 1600


# noinspection PyMethodMayBeStatic
class LauncherWindow(WindowWithTabs):
    @classmethod
    def open(cls, parent=None):
        return fsui.open_window_instance(cls, parent)

    @classmethod
    def current(cls):
        return fsui.current_window_instance(cls)

    def __init__(self, fsgs=None):
        LauncherWindow._current = weakref.ref(self)
        # Old name
        if fsgs is not None:
            self.fsgs = fsgs
        else:
            self.fsgs = default_context()
        # New name
        self.gsc = self.fsgs

        from launcher.launcherapp import LauncherApp

        LauncherApp.pre_start()

        border = True
        maximize = None
        menu = False
        if launcher.ui.get_screen_size()[1] <= 768:
            if fstd.desktop.is_running_gnome_3():
                border = False
                if "--window-border" in sys.argv:
                    border = True
            maximize = True
        if Skin.fws():
            border = False
            menu = True
        if "--no-window-border" in sys.argv:
            border = False

        if fsgs_module.product == "OpenRetro":
            app_name = "OpenRetro Launcher"
        else:
            app_name = "FS-UAE Launcher"
        if Skin.fws():
            title = app_name
        else:
            title = "{} {}".format(app_name, Application.instance().version)

        WindowWithTabs.__init__(self, None, title, border=border, menu=menu)
        icon = self.find_icon()
        if icon:
            self.set_icon_from_path(icon)
        self.main_layout = fsui.VerticalLayout()

        self.books = []
        self.config_browser = None
        self.configurations_panel = None
        self.main_menu_close_time = 0
        self.main_panel = None
        self._menu = None
        self.menu_button = None
        self.quick_settings_panel = None
        self.tab_panels = []
        self.user_button = None
        self.user_menu_close_time = 0

        self.set_content(self.main_layout)
        self.column_layout = fsui.HorizontalLayout()
        self.main_layout.add(self.column_layout, fill=True, expand=True)

        # left border
        self.create_column(
            0, min_width=Skin.get_window_padding_left(), content=False
        )
        # left content
        # if fs_uae_launcher.ui.get_screen_size()[0] > 1024:
        #     left_width = 518
        # else:
        #     left_width = 400
        # FIXME: adjust
        left_width = 462
        self.create_column(1, min_width=left_width)

        # right content
        # right_width = Constants.SCREEN_SIZE[0] * 2 + 21 + 10 + 10
        extra_screenshot_width = Constants.SCREEN_SIZE[0] + 21
        need_width = 1280

        right_width = 518

        if self.is_editor_enabled():
            # need_width += extra_screen_width
            pass
        else:
            if launcher.ui.get_screen_size()[0] >= need_width:
                # make room for one more screenshot
                right_width += extra_screenshot_width
                pass

        # right_width = 518

        # if Skin.EXTRA_GROUP_MARGIN:
        #     self.main_layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)

        self.create_column(2, min_width=right_width, expand=True)

        # right border
        self.create_column(
            3, min_width=Skin.get_window_padding_right(), content=False
        )

        # if self.is_editor_enabled():
        #     from ..editor.EditorGroup import EditorGroup
        #     editor = EditorGroup(self)
        #     self.main_layout.add(editor, fill=True, expand=True,
        #                          margin_right=20)

        # if fs_uae_launcher.ui.get_screen_size()[1] >= 768:
        #     right_margin = 0

        if launcher.ui.get_screen_size()[1] >= 720:
            self.bottom_layout = fsui.HorizontalLayout()
            self.main_layout.add(self.bottom_layout, fill=True)

            bottom_panel = BottomPanel(self)
            bottom_panel.set_min_width(10)
            bottom_panel.set_min_height(Skin.get_bottom_panel_height())
            self.bottom_layout.add(bottom_panel, fill=True)

            self.screenshots_panel = ScreenshotsPanel(self)
            self.bottom_layout.add(self.screenshots_panel, fill=True)

            self.game_info_panel = GameInfoPanel(self)
            self.game_info_panel.set_min_width(GAME_INFO_PANEL_MIN_WIDTH)
            self.bottom_layout.add(
                self.game_info_panel, fill=True, expand=True
            )

            bottom_panel = BottomPanel(self)
            bottom_panel.set_min_width(10)
            bottom_panel.set_min_height(Skin.get_bottom_panel_height())
            self.bottom_layout.add(bottom_panel, fill=True)
        else:
            self.bottom_layout = None
            self.screenshots_panel = None

        # right_margin = -10 - Skin.EXTRA_GROUP_MARGIN
        #     else:
        #         bottom_panel = None
        #     # FIXME:
        #     if bottom_panel is None:
        #         layout.add_spacer(0, Skin.get_bottom_panel_height())
        #     else:
        #         bottom_panel.set_min_height(Skin.get_bottom_panel_height())
        #         layout.add(bottom_panel, fill=True, margin_right=right_margin)
        # elif column == 1:
        #     group = LaunchGroup(self)
        #     layout.add(group, fill=True, margin=10, margin_top=0)
        #     layout.add_spacer(0, 10)

        self.realize_tabs()
        # self._menu = self.create_menu()
        if fsui.System.macosx and fsui.toolkit == "wx":
            # import wx
            # self.tools_menu = self.create_menu()
            # menu_bar = wx.MenuBar()
            # menu_bar.Append(self.tools_menu._menu, _("Tools"))
            # self.SetMenuBar(menu_bar)
            pass

        self.status_bar = StatusBar(self)
        self.layout.add(self.status_bar, fill=True)

        was_maximized = LauncherSettings.get("maximized") == "1"

        if LauncherSettings.get(Option.LAUNCHER_CONFIG_FEATURE) == "1":
            if launcher.ui.get_screen_size()[0] > 1300:
                if launcher.ui.get_screen_size()[1] > 1000:
                    self.layout.min_width = 1280
                    self.layout.min_height = 720

        self.set_size(self.layout.get_min_size())

        self.center_on_screen()
        if was_maximized or maximize:
            self.maximize()

        LauncherSignal.add_listener("scan_done", self)
        LauncherSignal.add_listener("setting", self)

        self.update_title()
        self.check_for_update_once()
        self.implicit_config_handler = ImplicitConfigHandler(self)

    def on_destroy(self):
        # FIXME: Is this being run?
        print("LauncherWindow.on_destroy")
        LauncherSignal.remove_listener("scan_done", self)
        LauncherSignal.remove_listener("setting", self)

    checked_for_update = False

    @classmethod
    def check_for_update_once(cls):
        if cls.checked_for_update:
            return
        UpdateManager.run_update_check()
        cls.checked_for_update = True

    def find_icon(self):
        icon = None

        def check_icon(path):
            path = os.path.join(path, "fs-uae-launcher.ico")
            if os.path.exists(path):
                return path
            return None

        if not icon:
            icon = check_icon("share/fs-uae-launcher")
        if not icon:
            icon = check_icon("launcher/share/fs-uae-launcher")

        # FIXME: should check data directories (XDG_DATA_DIRS) properly
        # instead
        if not icon:
            # FIXME: this encoding / decoding is a bit ridiculous, but, this is
            # for Python 2.x..
            icon = check_icon(
                os.path.expanduser(
                    "~/.local/share/fs-uae-launcher".encode(
                        sys.getfilesystemencoding()
                    )
                ).decode(sys.getfilesystemencoding())
            )

        # FIXME: Check relative to executable instead
        if not icon:
            icon = check_icon("/usr/local/share/fs-uae-launcher")
        if not icon:
            icon = check_icon("/usr/share/fs-uae-launcher")

        if sys.platform == "darwin":
            # Icons come from the app bundles
            icon = None
        return icon

    def on_scan_done_signal(self):
        print("LauncherWindow.on_scan_done_signal")
        LauncherConfig.update_kickstart()

    def update_title(self):
        if Skin.fws():
            # Just want to keep "FS-UAE Launcher" in the title
            return
        auth = LauncherSettings.get(Option.DATABASE_AUTH)
        if auth:
            username = LauncherSettings.get(Option.DATABASE_USERNAME)
            login_info = username
        else:
            login_info = gettext("Not logged in")
        if fsgs_module.product == "OpenRetro":
            app_name = "OpenRetro Launcher"
        else:
            app_name = "FS-UAE Launcher"
        title = "{} {} ({})".format(
            app_name, Application.instance().version, login_info
        )
        self.set_title(title)

    def on_setting(self, key, value):
        unused(value)
        if key == Option.DATABASE_AUTH:
            self.update_title()

    def on_resize(self):
        width, height = self.get_size()
        # print("on_resize, size =", width, height)
        if self.is_maximized():
            if LauncherSettings.get("maximized") != "1":
                LauncherSettings.set("maximized", "1")
        else:
            if LauncherSettings.get("maximized") != "0":
                LauncherSettings.set("maximized", "0")
        if self.screenshots_panel is not None:
            available = width - 2 * 10 - self.game_info_panel.get_min_width()
            num_screenshots = int(
                (available - 22 + 22) / (Constants.SCREEN_SIZE[0] + 22)
            )
            # print("Screenshots count:", num_screenshots)
            screenshots_panel_width = (
                (Constants.SCREEN_SIZE[0] + 22) * num_screenshots - 22 + 22
            )
            self.screenshots_panel.set_min_width(screenshots_panel_width)
        if width > CONFIGURATIONS_PANEL_WIDTH_THRESHOLD_2:
            configurations_width = (
                SCREENSHOT_WIDTH * 3 + SCREENSHOT_SPACING * 2 + MARGIN
            )
            self.configurations_panel.set_min_width(
                CONFIGURATIONS_PANEL_WIDTH_2
            )
        else:
            self.configurations_panel.set_min_width(
                CONFIGURATIONS_PANEL_WIDTH_1
            )
        super().on_resize()

    def is_editor_enabled(self):
        return "--editor" in sys.argv

    def create_column(self, column, content=True, expand=False, min_width=0):
        layout = fsui.VerticalLayout()
        self.column_layout.add(layout, fill=True, expand=expand)
        if min_width:
            layout.add_spacer(min_width, 0)

        layout.add_spacer(0, 10 + Skin.EXTRA_GROUP_MARGIN)

        if content:
            book = Book(self)
            Skin.set_background_color(book)
            if column == 1:
                margin_right = Skin.EXTRA_GROUP_MARGIN
                # expand = False
                expand = True
            else:
                margin_right = 0
                expand = True
            hor_layout = fsui.HorizontalLayout()
            layout.add(
                hor_layout, fill=True, expand=expand, margin_right=margin_right
            )
            # hor_layout.add(
            #     book, fill=True, expand=expand, margin_right=margin_right)

            vert_layout = fsui.VerticalLayout()
            hor_layout.add(
                vert_layout,
                fill=True,
                expand=expand,
                margin_right=margin_right,
            )

            hor_layout_2 = fsui.HorizontalLayout()
            vert_layout.add(hor_layout_2, fill=True, expand=True)

            # vert_layout.add(book, fill=True, expand=True)
            hor_layout_2.add(book, fill=True, expand=True)

            #                        self.add_tab_panel(LaunchGroup, expand=False)

            # if column == 2 and \
            #         LauncherSettings.get(
            #             Option.LAUNCHER_CONFIG_FEATURE) == "1":
            #     if launcher.ui.get_screen_size()[0] >= 1280:
            #         hor_layout.add_spacer(10)
            #         # if not Skin.fws():
            #         #     line_panel = fsui.Panel(self)
            #         #     line_panel.set_background_color(
            #         #         fsui.Color(*BORDER_COLOR))
            #         #     line_panel.set_min_width(2)
            #         #     hor_layout.add(
            #         #         line_panel, fill=True, margin_top=-10,
            #         #                    margin_bottom=-10)
            #         self.config_browser = ConfigBrowser(self)
            #         self.config_browser.set_min_width(CONFIG_BROWSER_WIDTH)
            #         # if Skin.fws():
            #         hor_layout.add(self.config_browser, fill=True,
            #                        expand=0.5, margin=-10, margin_left=0)
            #         # else:
            #         #     hor_layout.add(self.config_browser, fill=True,
            #         #                    expand=0.5, margin=10)

            if column == 2:
                # hor_layout.add_spacer(10)
                self.quick_settings_panel = QuickSettingsPanel(self, self.fsgs)
                self.quick_settings_panel.set_min_width(QUICK_SETTINGS_WIDTH)
                # hor_layout.add(self.quick_settings_panel, fill=True,
                #                expand=0.5, margin=-10, margin_left=0)
                # hor_layout.add(self.quick_settings_panel, fill=True,
                #                expand=0.5, margin=0, margin_top=0)

                # Adding to hor_layout_2 will make it appear "one level"
                # further in.
                # hor_layout_2.add(self.quick_settings_panel, fill=True,
                #                  margin=0, margin_top=0)
                if LauncherSettings.get(Option.QUICK_SETTINGS) != "1":
                    self.quick_settings_panel.hide()

                hor_layout.add(
                    self.quick_settings_panel,
                    fill=True,
                    margin=0,
                    margin_top=0,
                )

            if column == 2:
                # if fs_uae_launcher.ui.get_screen_size()[1] >= 1024:
                #     vert_layout.add_spacer(100)

                hor2_layout = fsui.HorizontalLayout()
                vert_layout.add(hor2_layout, fill=True, margin=10)
                launch_group = LaunchGroup(self, self.gsc)
                # hor2_layout.add_spacer(0, expand=True)
                hor2_layout.add(launch_group, expand=True)

            self.books.append(book)
            self.add_column_content(column)
        else:
            layout.add_spacer(0, expand=True)
            self.books.append(None)

        # if column == 1 and Settings.get(Option.CONFIG_FEATURE) == "1":
        #     from fs_uae_launcher.ui.config.browser import ConfigBrowser
        #     config_browser = ConfigBrowser(self)
        #     config_browser.set_min_height(200)
        #     layout.add(config_browser, fill=True, expand=True, margin=10)

        layout.add_spacer(0, 10 + Skin.EXTRA_GROUP_MARGIN)
        # if fs_uae_launcher.ui.get_screen_size()[1] >= 768:
        #     right_margin = 0
        #     if column == 0:
        #         bottom_panel = BottomPanel(self)
        #     elif column == 2:
        #         bottom_panel = GameInfoPanel(self)
        #     elif column == 1:
        #         bottom_panel = ScreenshotsPanel(self)
        #         right_margin = -10 - Skin.EXTRA_GROUP_MARGIN
        #     else:
        #         bottom_panel = None
        #     # FIXME:
        #     if bottom_panel is None:
        #         layout.add_spacer(0, Skin.get_bottom_panel_height())
        #     else:
        #         bottom_panel.set_min_height(Skin.get_bottom_panel_height())
        #         layout.add(bottom_panel, fill=True, margin_right=right_margin)
        # elif column == 1:
        #     group = LaunchGroup(self)
        #     layout.add(group, fill=True, margin=10, margin_top=0)
        #     layout.add_spacer(0, 10)

    def add_column_content(self, column):
        default_page_index = 0
        default_tab_index_offset = 0
        if column == 1:
            if Skin.fws():
                self.add_tab_spacer(5)
                self.add_tab_spacer(64)
            else:
                # if USE_MAIN_MENU:
                icon = fsui.Image("launcher:res/32/main-menu.png")
                self.menu_button = self.add_tab_button(
                    None,
                    icon,
                    gettext("Main Menu"),
                    menu_function=self.open_main_menu,
                    left_padding=5,
                )
            # default_tab_index_offset = 1
            # self.add_tab_spacer(60)
            # else:
            #    self.add_tab_spacer(10)

            # page_index = 0
            self.configurations_panel = self.add_page(
                column,
                ConfigurationsPanel,
                "",
                gettext("Configurations"),
                gettext("Configuration Browser"),
            )

            self.add_tab_spacer(0, expand=True)

        elif column == 2:
            self.new_tab_group()

            self.add_tab_spacer(10)
            self.add_page(
                column,
                MainPanel,
                "32x32/go-home",
                gettext("Config"),
                gettext("Main Configuration Options"),
            )
            self.add_page(
                column,
                FloppiesPanel,
                "32x32/media-floppy",
                gettext("Floppies"),
                gettext("Floppy Drives"),
            )
            self.add_page(
                column,
                CDPanel,
                "32x32/media-optical",
                gettext("CD-ROMs"),
                gettext("CD-ROM Drives"),
            )
            # noinspection SpellCheckingInspection
            self.add_page(
                column,
                HardDrivesPanel,
                "32x32/drive-harddisk",
                gettext("Hard Drives"),
            )
            self.add_scroll_page(
                column,
                RomRamPanel,
                "32x32/application-x-firmware",
                gettext("Hardware"),
                gettext("ROM and RAM"),
            )
            self.add_page(
                column,
                InputPanel,
                "32x32/applications-games",
                gettext("Input"),
                gettext("Input Options"),
            )
            self.add_scroll_page(
                column,
                ExpansionsPanel,
                "32x32/audio-card",
                gettext("Expansions"),
                gettext("Expansions"),
            )
            self.add_scroll_page(
                column,
                AdditionalConfigPanel,
                "32x32/system-shutdown",
                gettext("Additional Configuration"),
                gettext("Additional Configuration"),
            )

            self.add_tab_spacer(0, expand=True)
            self.add_tab_spacer(10)
            self.add_tab_button(
                self.toggle_quick_settings_sidebar,
                fsui.Image("launcher:res/32/quick-settings-sidebar.png"),
                gettext("Toggle quick settings sidebar"),
                right_padding=5,
            )

            # column - 1 is the group id of the tab group
            self.select_tab(
                default_page_index + default_tab_index_offset, column - 1
            )
        self.books[column].set_page(default_page_index)

    def create_menu(self):
        menu = fsui.PopupMenu()

        if LauncherSettings.get(Option.DATABASE_AUTH):
            menu.add_item(
                gettext("Update Game Database"), self.on_update_game_database
            )
        menu.add_item(
            gettext("Update File Database"), self.on_update_file_database
        )

        menu.add_separator()

        menu.add_item(gettext("ADF Creator"), self.on_adf_creator)
        menu.add_item(gettext("HDF Creator"), self.on_hdf_creator)
        if LauncherSettings.get(Option.NETPLAY_FEATURE) != "0":
            menu.add_item(gettext("Net Play"), self.on_net_play)

        menu.add_separator()
        menu.add_item(
            gettext("Import Kickstarts") + "...", self.on_import_kickstarts
        )
        menu.add_item(
            gettext("Amiga Forever Import") + "...", self.on_import_kickstarts
        )

        menu.add_separator()
        self.add_user_menu_content(menu)

        menu.add_separator()
        # menu.add_item(
        #         gettext("Kickstarts & ROMs"), self.on_import_kickstarts)

        if LauncherSettings.get(Option.LAUNCHER_SETUP_WIZARD_FEATURE):
            menu.add_item(
                gettext("Setup Wizard") + "...", self.on_setup_wizard
            )

        menu.add_preferences_item(gettext("Settings"), self.on_settings_button)

        menu.add_separator()
        menu.add_item(
            gettext("About {name}").format(name="OpenRetro.org"),
            self.on_what_is_this,
        )
        if fsgs_module.product == "OpenRetro":
            app_name = "OpenRetro Launcher"
        else:
            app_name = "FS-UAE Launcher"
        menu.add_about_item(
            gettext("About {name}").format(name=app_name), self.on_about
        )

        menu.add_separator()
        menu.add_about_item(gettext("Quit"), self.on_quit)
        return menu

    def _add_page(self, book, instance, icon_name, title, tooltip):
        book.add_page(instance)
        if icon_name:
            icon = fsui.Image("launcher:res/{0}.png".format(icon_name))
        else:
            icon = None

        def function():
            book.set_page(instance)

        if icon:
            self.add_tab(function, icon, title, tooltip)
        return instance

    def add_page(self, column, content_class, icon_name, title, tooltip=""):
        book = self.books[column]
        instance = content_class(book)
        if content_class == MainPanel:
            self.main_panel = instance
        return self._add_page(book, instance, icon_name, title, tooltip)

    def add_scroll_page(
        self, column, content_class, icon_name, title, tooltip=""
    ):
        book = self.books[column]
        instance = ConfigScrollArea(book)
        content_instance = content_class(instance)
        instance.set_widget(content_instance)
        if content_class == MainPanel:
            self.main_panel = instance
        return self._add_page(book, instance, icon_name, title, tooltip)

    def on_custom_button(self):
        from .config.configdialog import ConfigDialog

        ConfigDialog.run(self.get_window())

    def menu(self):
        self._menu = self.create_menu()
        return self._menu

    def open_main_menu(self):
        if fsui.System.windows:
            if time.time() - getattr(self, "main_menu_close_time", 0) < 0.2:
                return
        self._menu = self.create_menu()
        if Skin.use_unified_toolbar():
            self.popup_menu(self._menu, (0, -2))
        else:
            self.menu_button.popup_menu(
                self._menu, (0, self.menu_button.size()[1] - 2)
            )
        if fsui.System.windows:
            self.main_menu_close_time = time.time()
        return self._menu

    def toggle_quick_settings_sidebar(self):
        print("Toggling settings sidebar")
        if self.quick_settings_panel.visible():
            self.quick_settings_panel.hide()
            LauncherSettings.set(Option.QUICK_SETTINGS, "0")
        else:
            self.quick_settings_panel.show()
            LauncherSettings.set(Option.QUICK_SETTINGS, "1")
        self.layout.update()

    # def open_user_menu(self):
    #     if fsui.System.windows:
    #         if time.time() - getattr(self, "user_menu_close_time", 0) < 0.2:
    #             return
    #     user_menu = self.create_user_menu()
    #     self.user_button.popup_menu(
    #         user_menu, (0, self.user_button.get_size()[1] - 2))
    #     if fsui.System.windows:
    #         self.user_menu_close_time = time.time()

    def add_user_menu_content(self, menu):
        if LauncherSettings.get(Option.DATABASE_AUTH):
            # menu.add_item(_("Log In / Register"), self.on_log_in)
            # menu.add_item(gettext("Update Game Database"),
            #               self.on_game_database_refresh)
            if is_locker_enabled():
                menu.add_item(
                    gettext("Upload Files to OpenRetro Locker"),
                    self.on_upload_locker_files,
                )
                # menu.add_separator()
            menu.add_item(gettext("Log Out"), self.on_log_out)
        else:
            menu.add_item(gettext("Log In / Register"), self.on_log_in)

    def create_user_menu(self):
        menu = fsui.Menu()
        self.add_user_menu_content(menu)
        return menu

    def on_log_in(self):
        print("on_log_in")
        LoginWindow.open(self)

    def on_log_out(self):
        print("on_log_out")
        LogoutWindow.open(self)

    def on_what_is_this(self):
        print("on_what_is_this")
        fstd.desktop.open_url_in_browser("https://openretro.org/about")

    def on_scan_button(self):
        from .scan import ScanDialog

        ScanDialog(self.get_window()).show()

    def on_settings_button(self):
        from launcher.settings.settings_dialog import SettingsDialog

        SettingsDialog.open(self)

    def on_setup_wizard(self):
        from launcher.setup.setupwizarddialog import SetupWizardDialog

        SetupWizardDialog.open(self)

    def on_about(self):
        dialog = AboutDialog(self.get_window())
        dialog.show()

    def on_quit(self):
        self.close()

    def on_import_kickstarts(self):
        KickstartImportDialog(self.get_window()).show()

    def on_update_file_database(self):
        self.on_scan_button()

    def on_update_game_database(self):
        RefreshWindow.open(self)

    def on_upload_locker_files(self):
        print("on_upload_locker_files")
        LockerUploaderWindow.open(self)

    def on_adf_creator(self):
        print("on_adf_creator")
        ADFCreatorWindow.open(self)

    def on_hdf_creator(self):
        print("on_hdf_creator")
        HDFCreatorWindow.open(self)

    def on_net_play(self):
        print("on_net_play")
        NetplayDialog.open()



from __future__ import absolute_import
from __future__ import print_function

import struct
import ctypes

from .block.BitmapBlock import BitmapBlock
from .block.BitmapExtBlock import BitmapExtBlock
from .DosType import *
from .FSError import *

class ADFSBitmap:
  def __init__(self, root_blk):
    self.root_blk = root_blk
    self.blkdev = self.root_blk.blkdev
    # state
    self.ext_blks = []
    self.bitmap_blks = []
    self.bitmap_data = None
    self.valid = False
    # bitmap block entries
    self.bitmap_blk_bytes = root_blk.blkdev.block_bytes - 4
    self.bitmap_blk_longs = root_blk.blkdev.block_longs - 1
    # calc size of bitmap
    self.bitmap_bits = self.blkdev.num_blocks - self.blkdev.reserved
    self.bitmap_longs = (self.bitmap_bits + 31) / 32
    self.bitmap_bytes = (self.bitmap_bits + 7) / 8
    # number of blocks required for bitmap (and bytes consumed there)
    self.bitmap_num_blks = (self.bitmap_longs + self.bitmap_blk_longs - 1) / self.bitmap_blk_longs
    self.bitmap_all_blk_bytes = self.bitmap_num_blks * self.bitmap_blk_bytes
    # blocks stored in root and in every ext block
    self.num_blks_in_root = len(self.root_blk.bitmap_ptrs)
    self.num_blks_in_ext = self.blkdev.block_longs - 1
    # number of ext blocks required
    self.num_ext = (self.bitmap_num_blks - self.num_blks_in_root + self.num_blks_in_ext - 1) / (self.num_blks_in_ext) 
    # start a root block
    self.find_start = root_blk.blk_num
  
  def create(self):
    # create data and preset with 0xff
    self.bitmap_data = ctypes.create_string_buffer(self.bitmap_all_blk_bytes)
    for i in xrange(self.bitmap_all_blk_bytes):
      self.bitmap_data[i] = chr(0xff)

    # clear bit for root block
    blk_pos = self.root_blk.blk_num
    self.clr_bit(blk_pos)
    blk_pos += 1
    
    # create ext blocks
    for i in xrange(self.num_ext):
      bm_ext = BitmapExtBlock(self.blkdev, blk_pos)
      bm_ext.create()
      self.clr_bit(blk_pos)
      blk_pos += 1
      self.ext_blks.append(bm_ext)
      
    # create bitmap blocks
    for i in xrange(self.bitmap_num_blks):
      bm = BitmapBlock(self.blkdev, blk_pos)
      bm.create()
      self.clr_bit(blk_pos)
      blk_pos += 1
      self.bitmap_blks.append(bm)
      
    # set pointers to ext blocks
    if self.num_ext > 0:
      self.root_blk.bitmap_ext_blk = self.ext_blks[0].blk_num
      for i in xrange(self.num_ext-1):
        bm_ext = self.ext_blks[i]
        bm_ext_next = self.ext_blks[i+1]
        bm_ext.bitmap_ext_blk = bm_ext_next.blk_num
    
    # set pointers to bitmap blocks
    cur_ext_index = 0
    cur_ext_pos = 0
    for i in xrange(self.bitmap_num_blks):
      blk_num = self.bitmap_blks[i].blk_num
      if i < self.num_blks_in_root:
        # pointers in root block
        self.root_blk.bitmap_ptrs[i] = blk_num
      else:
        # pointers in ext block
        self.ext_blks[cur_ext_index].bitmap_ptrs[cur_ext_pos] = blk_num
        cur_ext_pos += 1
        if cur_ext_pos == self.num_blks_in_ext:
          cur_ext_pos = 0
          cur_ext_index += 1
    self.valid = True
  
  def write(self):
    # write root block
    self.root_blk.write()
    # write ext blocks
    for ext_blk in self.ext_blks:
      ext_blk.write()
    self.write_only_bits()
      
  def write_only_bits(self):
    # For DOS6 and DOS7, the root block contains the number of free blocks_used
    # So we potentially have to update it even if only bits shall be rewritten
    if rootblock_tracks_used_blocks(self.root_blk.fstype):
      # TODO: Track the used blocks instead of recalculating it every time
      blocks_used = self.get_num_used()
      if blocks_used != self.root_blk.blocks_used:
        self.root_blk.blocks_used = blocks_used
        self.root_blk.write()
    
    # write bitmap blocks
    off = 0
    for blk in self.bitmap_blks:
      blk.set_bitmap_data(self.bitmap_data[off:off+self.bitmap_blk_bytes])
      blk.write()
      off += self.bitmap_blk_bytes
  
  def read(self):
    self.bitmap_blks = []
    bitmap_data = ""
    
    # get bitmap blocks from root block
    blocks = self.root_blk.bitmap_ptrs
    for blk in blocks:
      if blk == 0:
        break
      bm = BitmapBlock(self.blkdev, blk)
      bm.read()
      if not bm.valid:
        raise FSError(INVALID_BITMAP_BLOCK, block=bm)
      self.bitmap_blks.append(bm)
      bitmap_data += bm.get_bitmap_data()
      
    # now check extended bitmap blocks
    ext_blk = self.root_blk.bitmap_ext_blk
    while ext_blk != 0:
      bm_ext = BitmapExtBlock(self.blkdev, ext_blk)
      bm_ext.read()
      self.ext_blks.append(bm_ext)
      blocks = bm_ext.bitmap_ptrs
      for blk in blocks:
        if blk == 0:
          break
        bm = BitmapBlock(self.blkdev, blk)
        bm.read()
        if not bm.valid:
          raise FSError(INVALID_BITMAP_BLOCK, block=bm)
        bitmap_data += bm.get_bitmap_data()
        self.bitmap_blks.append(bm)
      ext_blk = bm_ext.bitmap_ext_blk

    # check bitmap data
    num_bm_blks = len(self.bitmap_blks)
    num_bytes = self.bitmap_blk_bytes * num_bm_blks
    if num_bytes != len(bitmap_data):
      raise FSError(BITMAP_SIZE_MISMATCH, node=self, extra="got=%d want=%d" % (len(bitmap_data), num_bytes))
    if num_bm_blks != self.bitmap_num_blks:
      raise FSError(BITMAP_BLOCK_COUNT_MISMATCH, node=self, extra="got=%d want=%d" % (self.bitmap_num_blks, num_bm_blks))

    # create a modyfiable bitmap
    self.bitmap_data = ctypes.create_string_buffer(bitmap_data)
    self.valid = True

  def find_free(self, start=None):
    # give start of search
    if start == None:
      pos = self.find_start
    else:
      pos = start
    # at most scan all bits
    num = self.bitmap_bits
    while num > 0:
      # a free bit?
      found = self.get_bit(pos)
      old_pos = pos
      pos += 1
      if pos == self.bitmap_bits + self.blkdev.reserved:
        pos = self.blkdev.reserved
      if found:
        # start a next position
        self.find_start = pos
        return old_pos
      num -= 1
    return None

  def find_n_free(self, num, start=None):
    first_blk = self.find_free(start)
    if first_blk == None:
      return None
    if num == 1:
      return [first_blk]
    result = [first_blk]
    for i in xrange(num-1):
      blk_num = self.find_free()
      if blk_num == None:
        return None
      if blk_num in result:
        return None
      result.append(blk_num)
    return result
    
  def get_num_free(self):
    num = 0
    for i in xrange(self.bitmap_bits):
      if self.get_bit(i):
        num+=1
    return num

  def get_num_used(self):
    num = 0
    for i in xrange(self.bitmap_bits):
      if not self.get_bit(i):
        num+=1
    return num
    
  def alloc_n(self, num, start=None):
    free_blks = self.find_n_free(num, start)
    if free_blks == None:
      return None
    for b in free_blks:
      self.clr_bit(b)
    self.write_only_bits()
    return free_blks

  def dealloc_n(self, blks):
    for b in blks:
      self.set_bit(b)
    self.write_only_bits()

  def get_bit(self, off):
    if off < self.blkdev.reserved or off >= self.blkdev.num_blocks:
      return None
    off = (off - self.blkdev.reserved)
    long_off = off / 32
    bit_off = off % 32
    val = struct.unpack_from(">I", self.bitmap_data, long_off * 4)[0]
    mask = 1 << bit_off
    return (val & mask) == mask

  # mark as free
  def set_bit(self, off):
    if off < self.blkdev.reserved or off >= self.blkdev.num_blocks:
      return False
    off = (off - self.blkdev.reserved)
    long_off = off / 32
    bit_off = off % 32
    val = struct.unpack_from(">I", self.bitmap_data, long_off * 4)[0]
    mask = 1 << bit_off
    val = val | mask
    struct.pack_into(">I", self.bitmap_data, long_off * 4, val)
    return True
  
  # mark as used
  def clr_bit(self, off):
    if off < self.blkdev.reserved or off >= self.blkdev.num_blocks:
      return False
    off = (off - self.blkdev.reserved)
    long_off = off / 32
    bit_off = off % 32
    val = struct.unpack_from(">I", self.bitmap_data, long_off * 4)[0]
    mask = 1 << bit_off
    val = val & ~mask
    struct.pack_into(">I", self.bitmap_data, long_off * 4, val)
    return True

  def dump(self):
    print("Bitmap:")
    print("  ext: ",self.ext_blks)
    print("  blks:",len(self.bitmap_blks))
    print("  bits:",len(self.bitmap_data) * 8,self.blkdev.num_blocks)
    
  def create_draw_bitmap(self):
    bm = ctypes.create_string_buffer(self.blkdev.num_blocks)
    for i in xrange(self.blkdev.num_blocks):
      bm[i] = chr(0)
    return bm
  
  def print_free(self, brief=False):
    bm = self.create_draw_bitmap()
    res = self.blkdev.reserved
    for i in xrange(self.blkdev.num_blocks):
      if i >= res and self.get_bit(i):
        bm[i] = 'F'
    self.print_draw_bitmap(bm, brief)
      
  def print_used(self, brief=False):
    bm = self.create_draw_bitmap()
    res = self.blkdev.reserved
    for i in xrange(self.blkdev.num_blocks):
      if i >= res and not self.get_bit(i):
        bm[i] = '#'
    self.print_draw_bitmap(bm, brief)
  
  def draw_on_bitmap(self, bm):
    # show reserved blocks
    res = self.blkdev.reserved
    bm[0:res] = "x" * res
    # root block
    bm[self.root_blk.blk_num] = 'R'
    # bitmap blocks
    for bm_blk in self.bitmap_blks:
      bm[bm_blk.blk_num] = 'b'
    # bitmap ext blocks
    for ext_blk in self.ext_blks:
      bm[ext_blk.blk_num] = 'B'

  def print_draw_bitmap(self, bm, brief=False):
    line = ""
    blk = 0
    blk_cyl = self.blkdev.sectors * self.blkdev.heads
    found = False
    for i in xrange(self.blkdev.num_blocks):
      c = bm[i]
      if ord(c) == 0:
        c = '.'
      else:
        found = True
      line += c
      if i % self.blkdev.sectors == self.blkdev.sectors - 1:
        line += " "
      if i % blk_cyl == blk_cyl - 1:
        if not brief or found:
          print("%8d: %s" % (blk,line))
        blk += blk_cyl
        line = ""
        found = False



#!/usr/bin/env python3

import sqlite3
import sys
import zlib
from binascii import hexlify


def main():
    databasepath = sys.argv[1]
    database = sqlite3.connect(databasepath)
    cursor = database.cursor()
    cursor.execute("SELECT uuid, data FROM game")
    print("{", end="")
    count = 0
    for row in cursor:
        s = hexlify(row[0]).decode("ASCII")
        uuid = f"{s[0:8]}-{s[8:12]}-{s[12:16]}-{s[16:20]}-{s[20:32]}"
        if row[1]:
            data = zlib.decompress(row[1])
            jsonstr = data.decode("UTF-8")
            print("," if count > 0 else "")
            print(f'    "{uuid}": {jsonstr}', end="")
            count += 1
    print("\n}")
    print(f"Exported {count} entries", file=sys.stderr)


if __name__ == "__main__":
    main()



import fsbc.system
import fsboot
import fsui
from fsbc.util import memoize

from ..launcher_settings import LauncherSettings
from ..option import Option
from .Constants import Constants

try:
    import workspace
except ImportError:
    workspace = None

# LEVEL = 0xce
LEVEL = 0xEB
# LEVEL = 0xe4


class LauncherTheme(object):
    __instance = None

    @classmethod
    def get(cls) -> "LauncherTheme":
        if cls.__instance is None:
            cls.__instance = cls()
        return cls.__instance

    def __init__(self):
        from fsui.qt import QPalette

        palette = QPalette()
        self.sidebar_list_background = fsui.Color(
            palette.color(QPalette.ColorRole.Base)
        )
        self.sidebar_list_row_height = 28
        self.sidebar_list_row_text = fsui.Color(
            palette.color(QPalette.ColorRole.HighlightedText)
        )
        self.sidebar_list_row_background = fsui.Color(
            palette.color(QPalette.ColorRole.Highlight)
        )

        self.text_color = fsui.Color(
            palette.color(QPalette.ColorRole.Text)
        )

        self.window_color = fsui.Color(
            palette.color(QPalette.ColorRole.Window)
        )
        self.dark_mode = self.window_color.red() < 128

        if Skin.fws():
            from workspace.ui.theme import WorkspaceTheme as WorkspaceTheme

            ws_theme = WorkspaceTheme.instance()
            self.sidebar_list_background = ws_theme.sidebar_background
            self.sidebar_list_row_text = fsui.Color(0xFF, 0xFF, 0xFF)
            self.sidebar_list_row_background = ws_theme.selection_background
        elif Skin.windows_10():
            self.sidebar_list_background = fsui.Color(0xE2, 0xE2, 0xE2)

    @property
    def has_close_buttons(self):
        return LauncherSettings.get(Option.LAUNCHER_CLOSE_BUTTONS) == "1"


class Skin(object):
    EXTRA_GROUP_MARGIN = 0
    _fws = None

    @classmethod
    def use_unified_toolbar(cls):
        return False

    @classmethod
    def get_background_color(cls):
        c = cls._get_background_color()
        if c is None:
            return None
        return c.copy()

    @classmethod
    @memoize
    def _get_background_color(cls):
        # FIXME: Rename to launcher_bg/background_color and document.
        value = LauncherSettings.get("ui_background_color")
        if len(value) == 7 and value[0] == "#":

            def convert(s):
                return int(s, 16)

            r = convert(value[1:3])
            g = convert(value[3:5])
            b = convert(value[5:7])
            return fsui.Color(r, g, b)
        if cls.windows_10():
            return None
        elif cls.fws():
            return None
        # elif fsbc.system.windows:
        #     # FIXME: Should really just check for Windows XP here, or maybe
        #     # just remove it altogether.
        #     return fsui.Color(LEVEL, LEVEL, LEVEL)
        # elif fsbc.system.macosx:
        #     return fsui.Color(237, 237, 237)
        return None

    @classmethod
    def set_background_color(cls, widget):
        color = cls.get_background_color()
        if color is not None:
            widget.set_background_color(cls.get_background_color())

    @classmethod
    def get_window_padding_left(cls):
        return 10 + cls.EXTRA_GROUP_MARGIN

    @classmethod
    def get_window_padding_right(cls):
        return 10 + cls.EXTRA_GROUP_MARGIN

    @classmethod
    def get_window_padding_bottom(cls):
        return 10 + cls.EXTRA_GROUP_MARGIN

    @classmethod
    def get_bottom_margin(cls):
        return 10 + cls.get_window_padding_bottom()

    @classmethod
    def get_bottom_panel_height(cls):
        return (
            Constants.SCREEN_SIZE[1] + 20 + 2 + 1 + 1 + cls.get_bottom_margin()
        )

    @classmethod
    def windows_10(cls):
        return fsui.theme == "fusion" and fsui.theme_variant == "windows10"

    @classmethod
    def adwaita(cls):
        return fsui.theme == "fusion" and fsui.theme_variant == "adwaita"

    @classmethod
    def fws(cls):
        if cls._fws is None:
            if fsboot.get("fws") == "1":
                cls._fws = True
            else:
                cls._fws = LauncherSettings.get(Option.LAUNCHER_THEME) == "fws"
            if workspace is None:
                cls._fws = None
        return cls._fws



import threading
import traceback
from typing import Any, Dict, List, Tuple
from weakref import ref

main_thread_id = threading.current_thread().ident
_signal_id = 0


def new_signal_id():
    global _signal_id
    _signal_id += 1
    return str(_signal_id)


class Listener(object):
    def __init__(self, signal, listener):
        self.description = "<Listener {0}>".format(listener)

        if signal == "config":
            # FIXME: temporary hack for legacy code
            listener = listener.on_config
        elif signal == "setting":
            # FIXME: temporary hack for legacy code
            listener = listener.on_setting
        try:
            # FIXME: temporary hack for legacy code
            listener = getattr(listener, "on_" + signal + "_signal")
        except AttributeError:
            pass

        if hasattr(listener, "__self__") and hasattr(listener, "__func__"):
            self.instance = ref(listener.__self__)
            self.function = ref(listener.__func__)
        else:
            self.instance = None
            self.function = ref(listener)

    def __call__(self, *args, **kwargs):
        if self.instance is not None:
            # noinspection PyCallingNonCallable
            instance = self.instance()
            function = self.function()
            if function is None or instance is None:
                return False
            function(instance, *args, **kwargs)
            return True
        else:
            function = self.function()
            if function is None:
                return False
            function(*args, **kwargs)
            return True

    def __repr__(self):
        return self.description


class Signal:
    # FIXME: should have type Dict[str, Callable]
    # or # type_xxx: Dict[str, Function]
    signal_listeners = {}  # type: Dict[str, Any]
    # listeners = {}
    notifications = []  # type: List[Tuple[str, Any]]
    lock = threading.Lock()

    quit = None  # type: Signal

    def __init__(self, signal=None):
        if not signal:
            signal = new_signal_id()
        self.signal = signal

    def connect(self, function):
        listener = Listener(self.signal, function)
        with self.lock:
            # if hasattr(listener, "__self__") and hasattr(listener,
            # "__func__"):
            #     listener = (listener.__self__, listener.__func__)

            # Signal.signal_listeners.setdefault(
            #     self.signal, []).append(ref(listener))
            Signal.signal_listeners.setdefault(self.signal, []).append(
                listener
            )

            # Signal.listeners.setdefault(listener, []).append(self.signal)

    def disconnect(self, function):
        listener = Listener(self.signal, function)
        with self.lock:
            listeners = Signal.signal_listeners[self.signal]
            # print("removing listener", listener)
            for i, v in enumerate(listeners):
                if (
                    v.function == listener.function
                    and v.instance == listener.instance
                ):
                    del listeners[i]
                    break

            if len(Signal.signal_listeners[self.signal]) == 0:
                del Signal.signal_listeners[self.signal]

            # Signal.listeners[listener].remove(self.signal)
            # if len(Signal.listeners[listener]) == 0:
            #     del Signal.listeners[listener]

    # @classmethod
    # def disconnect_all(cls, listener):
    #     with Signal.lock:
    #         for signal in Signal.listeners[listener]:
    #             Signal.signal_listeners[signal].remove(listener)
    #         del Signal.listeners[listener]

    def __call__(self, *args):
        self.notify(*args)

    def notify(self, *args):
        if threading.current_thread().ident == main_thread_id:
            Signal.process_signal(self.signal, *args)

        else:
            with Signal.lock:
                Signal.notifications.append((self.signal, args))
            # FIXME: this is just a hack, fsui.call_after should be
            # replaced with an Application-specified callback function
            try:
                fsui = __import__("fsui")
                fsui.call_after(Signal.process_all_signals)
            except AttributeError:
                pass

    @classmethod
    def process_all_signals(cls):
        with Signal.lock:
            if len(Signal.notifications) == 0:
                return
            notifications = Signal.notifications[:]
            Signal.notifications = []
        # print(notifications)
        for signal, args in notifications:
            remove_listeners = cls.process_signal(signal, *args)
            if len(remove_listeners) > 0:
                with Signal.lock:
                    for listener in remove_listeners:
                        print("FIXME: remove dead listener", listener)
                        Signal.signal_listeners[signal].remove(listener)
                    if len(Signal.signal_listeners[signal]) == 0:
                        del Signal.signal_listeners[signal]

    @classmethod
    def process_signal(cls, signal, *args):
        # print(self.listeners.setdefault(signal, []))
        remove_listeners = []
        for listener_ref in Signal.signal_listeners.setdefault(signal, []):
            # print(listener, signal, args)
            # listener = listener_ref()
            # if not listener:
            #     # FIXME: remove reference
            #     continue
            listener = listener_ref
            # print(signal, listener, listener.instance, listener.function)

            try:
                # instance = listener.instance() if listener.instance else None
                # function = listener.function()
                # print(signal, instance, function)
                # if function is None:
                #     # FIXME: remove listener / dead listener
                #     continue
                #
                # if instance is not None:
                #     if signal == "config":
                #         # FIXME: temporary hack for legacy code
                #         instance.on_config(*args)
                #         continue
                #     elif signal == "setting":
                #         # FIXME: temporary hack for legacy code
                #         instance.on_setting(*args)
                #         continue
                #     try:
                #         # FIXME: temporary hack for legacy code
                #         func = getattr(instance, "on_" + signal + "_signal")
                #     except AttributeError:
                #         pass
                #     else:
                #         func(*args)
                #         continue

                result = listener(*args)
                if not result:
                    remove_listeners.append(listener)
            except Exception:
                traceback.print_exc()
        return remove_listeners


Signal.quit = Signal("quit")



import os.path
import plistlib

#
# Example settings file for dmgbuild
#

# Use like this: dmgbuild -s settings.py "Test Volume" test.dmg

# You can actually use this file for your own application (not just TextEdit)
# by doing e.g.
#
#   dmgbuild -s settings.py -D app=/path/to/My.app "My Application" MyApp.dmg

# .. Useful stuff ..............................................................

application = defines.get("app", "/System/Applications/TextEdit.app")  # noqa: F821
appname = os.path.basename(application)


def icon_from_app(app_path):
    plist_path = os.path.join(app_path, "Contents", "Info.plist")
    with open(plist_path, "rb") as f:
        plist = plistlib.load(f)
    icon_name = plist["CFBundleIconFile"]
    icon_root, icon_ext = os.path.splitext(icon_name)
    if not icon_ext:
        icon_ext = ".icns"
    icon_name = icon_root + icon_ext
    return os.path.join(app_path, "Contents", "Resources", icon_name)


# .. Basics ....................................................................

# Uncomment to override the output filename
# filename = 'test.dmg'

# Uncomment to override the output volume name
# volume_name = 'Test'

# Volume format (see hdiutil create -help)
format = defines.get("format", "UDBZ")  # noqa: F821

# Compression level (if relevant)
# compression_level = 9

# Volume size
size = defines.get("size", None)  # noqa: F821

# Files to include
files = [application]

# Symlinks to create
symlinks = {"Applications": "/Applications"}

# Files to hide
# hide = [ 'Secret.data' ]

# Files to hide the extension of
# hide_extension = [ 'README.rst' ]

# Volume icon
#
# You can either define icon, in which case that icon file will be copied to the
# image, *or* you can define badge_icon, in which case the icon file you specify
# will be used to badge the system's Removable Disk icon. Badge icons require
# pyobjc-framework-Quartz.
#
# icon = '/path/to/icon.icns'
badge_icon = icon_from_app(application)

# Where to put the icons
icon_locations = {appname: (140, 120), "Applications": (500, 120)}

# .. Window configuration ......................................................

# Background
#
# This is a STRING containing any of the following:
#
#    #3344ff          - web-style RGB color
#    #34f             - web-style RGB color, short form (#34f == #3344ff)
#    rgb(1,0,0)       - RGB color, each value is between 0 and 1
#    hsl(120,1,.5)    - HSL (hue saturation lightness) color
#    hwb(300,0,0)     - HWB (hue whiteness blackness) color
#    cmyk(0,1,0,0)    - CMYK color
#    goldenrod        - X11/SVG named color
#    builtin-arrow    - A simple built-in background with a blue arrow
#    /foo/bar/baz.png - The path to an image file
#
# The hue component in hsl() and hwb() may include a unit; it defaults to
# degrees ('deg'), but also supports radians ('rad') and gradians ('grad'
# or 'gon').
#
# Other color components may be expressed either in the range 0 to 1, or
# as percentages (e.g. 60% is equivalent to 0.6).
background = "builtin-arrow"

show_status_bar = False
show_tab_view = False
show_toolbar = False
show_pathbar = False
show_sidebar = False
sidebar_width = 180

# Window position in ((x, y), (w, h)) format
window_rect = ((100, 100), (640, 280))

# Select the default view; must be one of
#
#    'icon-view'
#    'list-view'
#    'column-view'
#    'coverflow'
#
default_view = "icon-view"

# General view configuration
show_icon_preview = False

# Set these to True to force inclusion of icon/list view settings (otherwise
# we only include settings for the default view)
include_icon_view_settings = "auto"
include_list_view_settings = "auto"

# .. Icon view configuration ...................................................

arrange_by = None
grid_offset = (0, 0)
grid_spacing = 100
scroll_position = (0, 0)
label_pos = "bottom"  # or 'right'
text_size = 16
icon_size = 128

# .. List view configuration ...................................................

# Column names are as follows:
#
#   name
#   date-modified
#   date-created
#   date-added
#   date-last-opened
#   size
#   kind
#   label
#   version
#   comments
#
list_icon_size = 16
list_text_size = 12
list_scroll_position = (0, 0)
list_sort_by = "name"
list_use_relative_dates = True
list_calculate_all_sizes = (False,)
list_columns = ("name", "date-modified", "size", "kind", "date-added")
list_column_widths = {
    "name": 300,
    "date-modified": 181,
    "date-created": 181,
    "date-added": 181,
    "date-last-opened": 181,
    "size": 97,
    "kind": 115,
    "label": 100,
    "version": 75,
    "comments": 300,
}
list_column_sort_directions = {
    "name": "ascending",
    "date-modified": "descending",
    "date-created": "descending",
    "date-added": "descending",
    "date-last-opened": "descending",
    "size": "descending",
    "kind": "ascending",
    "label": "ascending",
    "version": "ascending",
    "comments": "ascending",
}



class Option(object):
    def __init__(self):
        self.priority = 0.0
        self.group = False
        self.title = "Option Title"
        self.subtitle = ""

    def activate(self):
        pass

    @classmethod
    def create_play_option(cls):
        return PlayOption()

    @classmethod
    def create_config_option(cls):
        return ConfigOption()

    @classmethod
    def create_screen_option(cls):
        return ScreenOption()

    @classmethod
    def create_aspect_option(cls):
        return AspectOption()

    @classmethod
    def create_group(cls, title, priority=0.0):
        option = Option()
        option.priority = priority
        option.title = title
        option.group = True
        return option


class PlayOption(Option):
    def __init__(self):
        Option.__init__(self)
        self.priority = -2.0
        self.title = "Play Game"
        self.subtitle = "Play Game"

    def activate(self):
        print("PlayItem.activate")
        return "PLAY"


class ConfigOption(Option):
    def __init__(self):
        Option.__init__(self)
        self.priority = -1.0
        self.title = "CONFIGURATION NAME"
        self.subtitle = "Choose Configuration"

    def activate(self):
        print("ChooseConfigItem.activate")
        return []


class AspectOption(Option):
    def __init__(self):
        Option.__init__(self)
        self.priority = 98.0
        self.title = "FILL ENTIRE DISPLAY"
        self.subtitle = "Display Configuration"

    def activate(self):
        print("AspectItem.activate")
        return []


class ScreenOption(Option):
    def __init__(self):
        Option.__init__(self)
        self.priority = 99.9
        self.title = "FULL-SCREEN 50HZ V"
        self.subtitle = "Screen Configuration"

    def activate(self):
        print("DisplayItem.activate")
        return []



from fsgs.drivers.messdriver import MessDriver


class MessMsxDriver(MessDriver):
    CONTROLLER = {
        "type": "gamepad",
        "description": "Gamepad",
        "mapping_name": "msx",
    }

    PORTS = [{"description": "1st Controller", "types": [CONTROLLER]}]

    def mess_configure(self):
        self.mess_configure_cartridge()

    def mess_input_mapping(self, _):
        return {
            "START": "P#_START",
            "MODE": "P#_SELECT",
            "UP": "P#_JOYSTICK_UP",
            "DOWN": "P#_JOYSTICK_DOWN",
            "LEFT": "P#_JOYSTICK_LEFT",
            "RIGHT": "P#_JOYSTICK_RIGHT",
            "A": "P#_BUTTON1",
            "B": "P#_BUTTON2",
            "C": "P#_BUTTON3",
            "X": "P#_BUTTON4",
            "Y": "P#_BUTTON5",
            "Z": "P#_BUTTON6",
        }

    def mess_romset(self):
        # if self.is_pal():
        #     return "msx", {}
        # else:
        return "msx", {}



import hashlib
import os
import zipfile
from urllib.parse import unquote
from uuid import NAMESPACE_URL, uuid5


class ArchiveUtil:
    TYPE_WHDLOAD = "TYPE_WHDLOAD"
    TYPE_DOS = "TYPE_DOS"

    def __init__(self, path):
        self.path = path
        self.extensions = set()
        self.ignore_extensions = set()
        self.ignore_comments = False

    def create_variant_uuid(self):
        return self.create_variant()["uuid"]

    def create_variant(self, type=None, file_callback=None):
        archive_name = os.path.basename(self.path)
        archive_name, archive_ext = os.path.splitext(archive_name)
        archive_ext = archive_ext.lower()
        if archive_ext == ".zip":
            class_ = zipfile.ZipFile
            # continue
        elif archive_ext == ".lha":
            from lhafile import Lhafile

            class_ = Lhafile
        else:
            print("skipping", archive_name)
            return

        # s = hashlib.sha1()
        # with open(path, "rb") as f:
        #     data = f.read()
        #     s.update(data)
        # archive_sha1 = s.hexdigest()
        variant = {}
        names = []
        icons = []
        lower_to_regular = {}
        archive = class_(self.path)

        for info in archive.infolist():
            name = info.filename
            _, ext = os.path.splitext(name)
            if ext and ext.lower() in self.ignore_extensions:
                continue
            if name.endswith(".uaem"):
                continue
            if "META-INF/" in name:
                continue

            # FIXME: For WHDLoad?
            # if name.startswith(archive_name + "/"):
            #     name = name[len(archive_name) + 1:]

            if info.comment and not self.ignore_comments:
                comment = info.comment
            else:
                comment = None
            try:
                metadata = archive.read(name + ".uaem").decode("UTF-8")
            except KeyError:
                pass
            else:
                # FIXME: Create a uaem metadata helper class
                # print(type(metadata))
                metadata = metadata.split("\n")[0].strip()
                # x_permissions = metadata[0:8]
                # x_date = metadata[9:31]
                comment = metadata[32:]
                comment = unquote(comment)

            lower_to_regular[name.lower()] = name

            if name.lower().endswith(".info"):
                if len(name.split("/")) == 2:
                    # We're at game folder level
                    info_data = archive.read(name)
                    info_data_lower = info_data.lower()
                    # if "whdload" in info_data_lower and \
                    #         "slave=" in info_data_lower:
                    if b"slave=" in info_data_lower:
                        # if ".slav" in info_data.lower():
                        # assert "slave" in info_data
                        icons.append(name)
                # All .info files are skipped, these are generally not identical
                # across installations, and are also not needed.
                continue

            # FIXME: use DOS charset instead...
            # name = name.decode("ISO-8859-1")

            if len(names) > 0:
                if names[-1][0].endswith("/"):
                    if name.startswith(names[-1][0]):
                        # Remove directory entry - we do not need that when
                        # the directory has content.
                        del names[len(names) - 1]
            names.append((name, name, comment))

        names.sort()
        file_sha1s = []
        file_list = []

        for dummy, name, comment in names:
            _, ext = os.path.splitext(name)
            if ext:
                self.extensions.add(ext.lower())
            # name_lower = name.lower()
            # content_hash.update(name.encode("ISO-8859-1"))

            if type == self.TYPE_WHDLOAD:
                list_name = "DH0/ " + name
            elif type == self.TYPE_DOS:
                list_name = "C/" + name.upper()
            else:
                list_name = name

            if list_name.endswith("/"):
                # Directory entry
                file_list.append({"name": list_name})
                continue

            # FIXME: encoding
            # data = archive.read(name.encode("ISO-8859-1"))
            data = archive.read(name)
            # content_hash.update(data)
            # data_hash.update(data)
            member_hash = hashlib.sha1()
            member_hash.update(data)
            file_sha1s.append(member_hash.hexdigest())
            file_list_item = {
                "name": list_name,
                "sha1": member_hash.hexdigest(),
                "size": len(data),
            }
            if comment:
                print(repr(comment))
                assert isinstance(comment, str)
                file_list_item["comment"] = comment
            file_list.append(file_list_item)
            if file_callback is not None:
                file_callback(sha1=member_hash.hexdigest(), data=data)

        url = "http://sha1.fengestad.no/" + "/".join(sorted(file_sha1s))
        variant["uuid"] = str(uuid5(NAMESPACE_URL, url))
        variant["file_list"] = file_list
        return variant



import platform
import sys
import traceback

from launcher.version import VERSION


def app_main():
    if "--help" in sys.argv:
        print(help_text)
        return
    print("FS-UAE Launcher {0}".format(VERSION))
    print(sys.argv)
    print(platform.uname())

    for i, arg in enumerate(sys.argv[:]):
        if arg == "--fullscreen" or arg == "--fullscreen=1":
            sys.argv[i] = "--settings:fullscreen=1"
        elif arg == "--no-fullscreen" or arg == "--fullscreen=0":
            sys.argv[i] = "--settings:fullscreen=0"

    from launcher.launcherapp import LauncherApp

    app = LauncherApp()
    if "--no-auto-detect-game" in sys.argv:
        sys.argv.remove("--no-auto-detect-game")
        LauncherApp.auto_detect_game = False
    try:
        app.start()
    except Exception as e:
        traceback.print_exc(file=sys.stderr)
        import fsui

        if "--no-gui" in sys.argv:
            pass
        else:
            fsui.show_error(
                "An error occurred starting FS-UAE Launcher:\n\n"
                + repr(e)
                + "\n\nFS-UAE Launcher cannot start "
                "because of this.",
                "FS-UAE Launcher",
            )
            fsui.show_error(traceback.format_exc())
    else:
        app.run()
        app.save_settings()

    # from fs_uae_launcher.netplay.IRC import IRC
    # IRC.stop()

    from fsbc.signal import Signal

    Signal("quit").notify()
    print("app_main done")


help_text = """\
FS-UAE Launcher {version}

Options:
  --new-config[=<platform>]

Options for directly launching games:
  --fullscreen
  --no-fullscreen
  --no-gui                             Do not show launch progress dialog
  --no-auto-detect-game                Do not auto-detect game from archive

TODO: Add more documentation

Or you can execute fs-uae-launcher <command> - where command is one of:
  * arcade                             Launch FS-UAE Arcade
  * fs-uae-netplay-server              Run netplay server
  * list-dirs                          List directories used by the Launcher
  * list-plugins                       List installed plugins
""".format(version=VERSION)



#!/usr/bin/env python2.7
# based heavily on "iops" by Benjamin Schweizer
# https://github.com/gopher/iops

from __future__ import absolute_import
from __future__ import print_function

import sys
import array
import struct
import os

def getblkdevsize(dev):
    """report the size for a block device"""
    size = 0
    if sys.platform == 'darwin':
        # mac os x ioctl from sys/disk.h
        import fcntl
        DKIOCGETBLOCKSIZE = 0x40046418  # _IOR('d', 24, uint32_t)
        DKIOCGETBLOCKCOUNT = 0x40086419 # _IOR('d', 25, uint64_t)

        fh = open(dev, 'r')
        buf = array.array('B', range(0,4))  # uint32
        r = fcntl.ioctl(fh.fileno(), DKIOCGETBLOCKSIZE, buf, 1)
        blocksize = struct.unpack('I', buf)[0]
        buf = array.array('B', range(0,8))  # uint64
        r = fcntl.ioctl(fh.fileno(), DKIOCGETBLOCKCOUNT, buf, 1)
        blockcount = struct.unpack('Q', buf)[0]
        fh.close()
        size = blocksize*blockcount

    elif sys.platform.startswith('freebsd'):
        # freebsd ioctl from sys/disk.h
        import fcntl
        DIOCGMEDIASIZE = 0x40086481 # _IOR('d', 129, uint64_t)

        fh = open(dev, 'r')
        buf = array.array('B', range(0,8))  # off_t / int64
        r = fcntl.ioctl(fh.fileno(), DIOCGMEDIASIZE, buf, 1)
        size = struct.unpack('q', buf)[0]
        fh.close()

    elif sys.platform == 'win32':
        # win32 ioctl from winioctl.h, requires pywin32
        try:
            import win32file
        except ImportError:
            raise SystemExit("Package pywin32 not found, see http://sf.net/projects/pywin32/")
        IOCTL_DISK_GET_DRIVE_GEOMETRY = 0x00070000
        dh = win32file.CreateFile(dev, 0, win32file.FILE_SHARE_READ, None, win32file.OPEN_EXISTING, 0, None)
        info = win32file.DeviceIoControl(dh, IOCTL_DISK_GET_DRIVE_GEOMETRY, '', 24)
        win32file.CloseHandle(dh)
        (cyl_lo, cyl_hi, media_type, tps, spt, bps) = struct.unpack('6L', info)
        size = ((cyl_hi << 32) + cyl_lo) * tps * spt * bps

    else: # linux or compat
        # linux 2.6 lseek from fcntl.h
        fh = open(dev, 'r')
        fh.seek(0,os.SEEK_END)
        size = fh.tell()
        fh.close()

    if not size:
        raise Exception("getblkdevsize: Unsupported platform")
    
    return size

# test
if __name__ == '__main__':
  for a in sys.argv[1:]:
    print(a, getblkdevsize(a))




import os
import shutil
from typing import Optional

from fsgs import Option
from fsgs.drivers.gamedriver import Emulator, GameDriver
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.input.mapper import InputMapper
from fsgs.plugins.pluginmanager import PluginManager
from fsgs.saves import SaveHandler


class RetroArchDriver(GameDriver):
    def __init__(self, fsgc, libretro_core, retroarch_state_dir):
        super().__init__(fsgc)
        self.emulator = Emulator("retroarch-fs")
        # self.emulator.allow_system_emulator = True
        # self.libretro_core = "no_core_specified"
        self.system_dir = self.temp_dir("system")
        self.save_handler = RetroArchSaveHandler(
            self.fsgc, options=self.options, emulator=retroarch_state_dir
        )
        # self.retroarch_state_dir = None

        self.libretro_core = libretro_core
        self.retroarch_state_dir = retroarch_state_dir
        # Dictionary with key-values which will be written to retroarch.cfg
        self.retroarch_config_file = self.temp_file("retroarch.cfg")
        self.retroarch_config = {}
        # FIXME
        self.retroarch_core_config = {}

    def prepare(self):
        self.erase_old_config()

        # if not os.path.exists(config_dir):
        #     os.makedirs(config_dir)
        # config_file = os.path.join(config_dir, "retroarch.cfg")

        # FIXME: Do not use /etc/retroarch.cfg as template.. how to prevent?

        self.save_handler.prepare()

        with open(self.retroarch_config_file.path, "w", encoding="UTF-8") as f:
            self.write_retroarch_config(f)
            self.write_retroarch_input_config(f)
            self.write_retroarch_video_config(f)

        self.emulator.args.append(
            "--appendconfig=" + self.retroarch_config_file.path
        )
        if self.use_fullscreen():
            self.emulator.args.append("--fullscreen")

        libretro_core = self.find_libretro_core(self.libretro_core)
        if not libretro_core:
            raise Exception(
                "Could not find libretro core {0!r}".format(self.libretro_core)
            )
        self.emulator.args.extend(["-L", libretro_core])

        # Verbose logging
        self.emulator.args.extend(["-v"])

    def run(self):
        with open(self.retroarch_config_file.path, "a", encoding="UTF-8") as f:
            for key, value in self.retroarch_config.items():
                f.write('{} = "{}"\n'.format(key, value))
        super().run()

    def finish(self):
        self.save_handler.finish()

    @staticmethod
    def find_libretro_core(name):
        # return "/usr/lib/x86_64-linux-gnu/libretro/{}.so".format(name)
        return PluginManager.instance().find_library_path(name)

    @staticmethod
    def find_retroarch_shader(name):
        # FIXME: Better to find data file based on path/provides rather than
        # hardcoding plugin name, but...
        plugin = PluginManager.instance().plugin("RetroArch-FS")
        return plugin.data_file_path("shaders/shaders_glsl/" + name + ".glslp")

    def display_aspect_ratio(self) -> Optional[float]:
        return 4 / 3

    def game_video_size(self):
        # FIXME: Dummy values
        print("[DRIVER] Warning: Using dummy game video size (320, 240)")
        return 320, 240

    def erase_old_config(self):
        config_dir = os.path.join(self.home.path, ".config", "retroarch")
        if os.path.exists(config_dir):
            shutil.rmtree(config_dir)

    def open_retroarch_core_options(self):
        config_dir = os.path.join(self.home.path, ".config", "retroarch")
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)
        config_file = os.path.join(config_dir, "retroarch-core-options.cfg")
        return open(config_file, "w", encoding="UTF-8")

    def write_retroarch_config(self, f):
        """
        joypad_autoconfig_dir = "~/.config/retroarch/autoconfig"
        """
        f.write(
            'screenshot_directory = "{}"\n'.format(
                FSGSDirectories.screenshots_output_dir()
            )
        )
        f.write('system_directory = "{}"\n'.format(self.system_dir.path))
        # noinspection SpellCheckingInspection

        f.write(
            'savefile_directory = "{}"\n'.format(
                self.save_handler.emulator_save_dir()
            )
        )
        f.write(
            'savestate_directory = "{}"\n'.format(
                self.save_handler.emulator_state_dir()
            )
        )

        # FIXME: Maybe enable autosave to save .srm while running the emulator
        # and not only on shutdown?
        # f.write("autosave_interval = 60\n")

        f.write("pause_nonactive = false\n")
        f.write("video_font_enable = false\n")
        f.write("rgui_show_start_screen = false\n")
        f.write("all_users_control_menu = true\n")
        f.write("video_gpu_screenshot = false\n")

        if self.g_sync():
            # Without timed frame limiting, there will be stuttering (probably)
            # due to (some) audio driver sync not being stable enough to give
            # a stable frame rate.

            # FIXME: Implement better G-SYNC method in RetroArch
            f.write("fastforward_ratio = 1.000000\n")

            # FIXME: It's possible the above "fix" would be better for
            # non-v-sync as well, if the audio sync is not stable.

            # f.write("audio_max_timing_skew = 0.0\n")
            # f.write("audio_rate_control_delta = 0.0\n")
            # f.write("video_refresh_rate = 144.0\n")
            # f.write("audio_sync = false\n")

        default_buffer_size = 40
        buffer_size = default_buffer_size
        if self.options[Option.RETROARCH_AUDIO_BUFFER]:
            try:
                buffer_size = int(self.options[Option.RETROARCH_AUDIO_BUFFER])
            except ValueError:
                print("WARNING: Invalid RetroArch audio buffer size specified")
            else:
                if buffer_size < 0 or buffer_size > 1000:
                    print("WARNING: RetroArch audio buffer size out of range")
                    buffer_size = default_buffer_size
        f.write("audio_latency = {}\n".format(buffer_size))

    def write_retroarch_input_config(self, f):
        f.write('input_driver = "sdl2"\n')
        f.write('input_enable_hotkey = "alt"\n')
        f.write('input_exit_emulator = "q"\n')
        f.write('input_toggle_fast_forward = "w"\n')
        f.write('input_screenshot = "s"\n')
        # f.write("input_toggle_fullscreen = \"enter\"\n")
        f.write('input_toggle_fullscreen = "f"\n')
        f.write('input_audio_mute = "m"\n')
        f.write('input_menu_toggle = "f12"\n')
        f.write('input_pause_toggle = "p"\n')

        for i, port in enumerate(self.ports):
            if port.device is None:
                continue
            input_mapping = self.retroarch_input_mapping(i)
            # FIXME: EXCLUDE DUPLICATE ITEMS IN INPUT MAPPING???
            mapper = RetroArchInputMapper(port, input_mapping)

            f.write('input_player1_joypad_index = "0"\n')
            f.write('input_player1_analog_dpad_mode = "0"\n')
            if port.device.type == "joystick":
                pass
            else:
                pass

            for key, value in mapper.items():
                # print("---->", key, value)
                postfix, value = value
                f.write('{}{} = "{}"\n'.format(key, postfix, value))
                postfixes = ["", "_btn", "_axis"]
                postfixes.remove(postfix)
                for postfix in postfixes:
                    f.write('{}{} = "{}"\n'.format(key, postfix, "nul"))

    def write_retroarch_video_config(self, f):
        # f.write("video_driver = \"gl\"\n")
        # FIXME
        if self.fullscreen_window_mode():
            f.write("video_windowed_fullscreen = true\n")
        else:
            f.write("video_windowed_fullscreen = false\n")

        # FIXME: 1 or 0?
        f.write("video_max_swapchain_images = 1\n")

        if self.effect() == self.CRT_EFFECT:
            video_shader = "crt/crt-aperture"
            video_scale = 2
        elif self.effect() == self.DOUBLE_EFFECT:
            if self.smoothing() == self.NO_SMOOTHING:
                video_shader = ""
            else:
                video_shader = "retro/sharp-bilinear-2x-prescale"
            video_scale = 2
        elif self.effect() == self.HQ2X_EFFECT:
            video_shader = "hqx/hq2x"
            video_scale = 2
        elif self.effect() == self.SCALE2X_EFFECT:
            video_shader = "scalenx/scale2x"
            video_scale = 2
        else:
            video_shader = ""
            video_scale = 1

        if self.smoothing() == self.NO_SMOOTHING or video_shader:
            f.write("video_smooth = false\n")

        if video_shader:
            print("[DRIVER] Video shader:", video_shader)
            video_shader_path = self.find_retroarch_shader(video_shader)
            print("[DRIVER] Video shader path:", video_shader_path)
            if video_shader_path:
                f.write('video_shader = "{}"\n'.format(video_shader_path))
                f.write("video_shader_enable = true\n")

        # FIXME: video_monitor_index = 0
        # FIXME: video_disable_composition = true
        if self.use_vsync():
            f.write("video_vsync = true\n")
        else:
            f.write("video_vsync = false\n")

        # aspect_ratio_index = 22
        # custom_viewport_width = 0
        # custom_viewport_height = 0
        # custom_viewport_x = 0
        # custom_viewport_y = 0

        if self.stretching() == self.STRETCH_ASPECT:
            aspect_ratio = self.display_aspect_ratio()
            if aspect_ratio is not None:
                f.write("aspect_ratio_index = 19\n")
                f.write("video_aspect_ratio = {:f}\n".format(aspect_ratio))
                # f.write("video_force_aspect = \"true\"\n")
                # f.write("video_aspect_ratio_auto = \"false\"\n")
        elif self.stretching() == self.STRETCH_FILL_SCREEN:
            screen_w, screen_h = self.screen_size()
            display_aspect = screen_w / screen_h
            f.write("aspect_ratio_index = 19\n")
            f.write("video_aspect_ratio = {:f}\n".format(display_aspect))
        else:
            f.write("aspect_ratio_index = 20\n")

        if self.scaling() == self.NO_SCALING:
            # FIXME: Window size rounding issues? E.g. 897x672 for 4:3
            # NES 3x display. Maybe set window size manually instead
            f.write("video_scale = {}\n".format(video_scale))

        # f.write("input_osk_toggle = \"nul\"\n")
        overlay_path = self.create_retroarch_layout()
        if overlay_path:
            f.write('input_overlay = "{}"\n'.format(overlay_path))
            f.write('input_overlay_opacity = "1.000000"\n')

    def retroarch_input_mapping(self, port):
        return {}

    def display_rect_fullscreen(self):
        # FIXME: Check square pixels option!

        screen_w, screen_h = self.screen_size()
        screen_aspect = screen_w / screen_h

        if self.stretching() == self.STRETCH_ASPECT:
            display_aspect = self.display_aspect_ratio()
        elif self.stretching() == self.STRETCH_FILL_SCREEN:
            screen_w, screen_h = self.screen_size()
            display_aspect = screen_w / screen_h
        else:
            game_w, game_h = self.game_video_size()
            display_aspect = game_w / game_h

        # FIXME: round to nearest multiple of two?
        if screen_aspect >= display_aspect:
            game_h = screen_h
            game_w = round(game_h * display_aspect)
        else:
            game_w = screen_w
            game_h = round(game_w / display_aspect)
        game_x = round((screen_w - game_w) / 2)
        game_y = round((screen_h - game_h) / 2)
        return game_x, game_y, game_w, game_h

    def create_retroarch_layout(self):
        if self.stretching() == self.STRETCH_FILL_SCREEN or not self.bezel():
            return
        if not self.use_fullscreen():
            return

        # FIXME: file cmp?
        paths = self.prepare_emulator_skin()
        print(paths)

        # FIXME: SUPPORT frame = 0 ( bezel = 0) option

        # FIXME: With no bezel, we should still use a black bezel to
        # hide screen stretching

        screen_width, screen_height = self.screen_size()
        # dst_x = 0
        dst_y = 0
        # dst_w = 0
        # dst_w = 160
        dst_h = screen_height

        # Bezel size is normalized against 1080 (height)
        scale = screen_height / 1080
        # Bezel width: 160
        dst_w = round(160 * scale)

        game_x, game_y, game_w, game_h = self.display_rect_fullscreen()

        from fsui.qt import QImage, QPainter, QRect, QSize, Qt

        image = QImage(
            QSize(screen_width, screen_height), QImage.Format_RGBA8888
        )
        image.fill(Qt.transparent)
        # painter = image.paintEngine()
        painter = QPainter(image)

        dst_x = game_x - dst_w
        left = QImage(paths["left"])
        painter.drawImage(QRect(dst_x, dst_y, dst_w, dst_h), left)

        dst_x = game_x + game_w
        right = QImage(paths["right"])
        painter.drawImage(QRect(dst_x, dst_y, dst_w, dst_h), right)
        painter.end()

        overlay_png_file = self.temp_file("overlay.png").path
        image.save(overlay_png_file)

        # noinspection SpellCheckingInspection
        overlay_config = """overlays = 1
overlay0_overlay = {overlay}
overlay0_full_screen = true
overlay0_rect = "0.0,0.0,1.0,1.0"
overlay0_descs = 0
""".format(overlay=overlay_png_file)
        #         overlay_config = (
        #             """overlays = 2
        # overlay0_overlay = {left}
        # overlay0_full_screen = true
        # overlay0_rect = "0.0,0.0,0.12,1.0"
        # overlay0_descs = 0
        # overlay1_overlay = {right}
        # overlay1_full_screen = true
        # overlay1_rect = "0.8,0.0,0.2,1.0"
        # overlay1_descs = 0
        #
        # """.format(left=paths["left"], right=paths["right"]))
        overlay_config_file = self.temp_file("overlay.cfg")
        with open(overlay_config_file.path, "w") as f:
            f.write(overlay_config)
        return overlay_config_file.path

    def window_size(self):
        return 0, 0


class RetroArchInputMapper(InputMapper):
    def __init__(self, port, mapping):
        super().__init__(port, mapping, multiple=False)

    def axis(self, axis, positive):
        dir_str = "+" if positive else "-"
        return "_axis", dir_str + str(axis)

    def hat(self, hat, direction):
        return "_btn", "h{}{}".format(hat, direction)

    def button(self, button):
        return "_btn", button

    def key(self, key):
        # FIXME: Correct RetroArch key names
        name = key.sdl_name[5:].lower()
        if name == "return":
            # FIXME: HACK
            name = "enter"
        return "", name


class RetroArchSaveHandler(SaveHandler):
    def __init__(self, fsgc, options, emulator):
        super().__init__(fsgc, options, emulator=emulator)

    def prepare(self):
        super().prepare()

    def finish(self):
        super().finish()



import fsui
from launcher.i18n import gettext
from launcher.option import Option
from launcher.settings.settings_page import SettingsPage


class GameDatabaseSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("database-settings", "pkg:workspace")
        gettext("Game Database Settings")
        title = gettext("Game Database")
        subtitle = ""
        self.add_header(icon, title, subtitle)

        self.hori_layout = None
        self.hori_counter = 0

        self.add_option(Option.DATABASE_SHOW_GAMES)
        self.add_option(Option.DATABASE_SHOW_ADULT)
        self.add_option(Option.DATABASE_SHOW_UNPUBLISHED)

    #     if openretro or settings.get(Option.PLATFORMS_FEATURE) == "1":
    #         self.add_section(gettext("Additional Databases"))
    #
    #         label = fsui.MultiLineLabel(
    #             self, gettext(
    #                 "Note: This is an experimental feature. "
    #                 "Additional plugins are needed."), 640)
    #         self.layout.add(label, margin_top=20, margin_bottom=20)
    #
    #         self.add_database_option(Option.CPC_DATABASE, "Amstrad CPC")
    #         self.add_database_option(Option.ARCADE_DATABASE, "Arcade")
    #         self.add_database_option(Option.ATARI_DATABASE, "Atari ST")
    #         self.add_database_option(Option.C64_DATABASE, "Commodore 64")
    #         self.add_database_option(Option.DOS_DATABASE, "DOS")
    #         self.add_database_option(Option.GB_DATABASE, "Game Boy")
    #         self.add_database_option(Option.GBA_DATABASE, "Game Boy Advance")
    #         self.add_database_option(Option.GBC_DATABASE, "Game Boy Color")
    #         self.add_database_option(Option.NES_DATABASE, "Nintendo")
    #         self.add_database_option(Option.PSX_DATABASE, "PlayStation")
    #         self.add_database_option(Option.SNES_DATABASE, "Super Nintendo")
    #         self.add_database_option(Option.ZXS_DATABASE, "ZX Spectrum")
    #
    #         # label = fsui.MultiLineLabel(
    #         #     self, gettext(
    #         #         "Note: Support for additional game databases is an "
    #         #         "experimental feature and does not provide the "
    #         #         "same level of maturity as Amiga/CDTV/CD32. "
    #         #         "Also, additional plugins are needed to play the "
    #         #         "games."), 640)
    #         # self.layout.add(label, margin_top=20)
    #
    # def add_database_option(self, name, description=""):
    #     self.options_on_page.add(name)
    #     group = OptionUI.create_group(
    #         self, name, description=description, help_button=False)
    #
    #     if self.hori_counter % 2 == 0:
    #         self.hori_layout = fsui.HorizontalLayout()
    #         self.layout.add(
    #             self.hori_layout, fill=True, margin_top=10, margin_bottom=10,
    #             margin_left=-10, margin_right=-10)
    #     self.hori_layout.add(group, fill=True, expand=-1, margin=10,
    #                          margin_top=0, margin_bottom=0)
    #     if self.hori_counter % 2 == 0:
    #         self.hori_layout.add_spacer(10)
    #     self.hori_counter += 1
